"""This module defines a Hamiltonian."""

import functools

import numpy as np
from ase.units import Ha

from gpaw.arraydict import ArrayDict
from gpaw.external import create_external_potential
from gpaw.lfc import LFC
from gpaw.poisson import PoissonSolver
from gpaw.spinorbit import soc
from gpaw.transformers import Transformer
from gpaw.utilities import (pack2, pack_atomic_matrices, unpack,
                            unpack_atomic_matrices)
from gpaw.utilities.partition import AtomPartition

ENERGY_NAMES = ['e_kinetic', 'e_coulomb', 'e_zero', 'e_external', 'e_xc',
                'e_entropy', 'e_total_free', 'e_total_extrapolated']


def apply_non_local_hamilton(dH_asp, collinear, P, out=None):
    if out is None:
        out = P.new()
    for a, I1, I2 in P.indices:
        if collinear:
            dH_ii = unpack(dH_asp[a][P.spin])
            out.array[:, I1:I2] = np.dot(P.array[:, I1:I2], dH_ii)
        else:
            dH_xp = dH_asp[a]
            # We need the transpose because
            # we are dotting from the left
            dH_ii = unpack(dH_xp[0]).T
            dH_vii = [unpack(dH_p).T for dH_p in dH_xp[1:]]
            out.array[:, 0, I1:I2] = (np.dot(P.array[:, 0, I1:I2],
                                             dH_ii + dH_vii[2]) +
                                      np.dot(P.array[:, 1, I1:I2],
                                             dH_vii[0] - 1j * dH_vii[1]))
            out.array[:, 1, I1:I2] = (np.dot(P.array[:, 1, I1:I2],
                                             dH_ii - dH_vii[2]) +
                                      np.dot(P.array[:, 0, I1:I2],
                                             dH_vii[0] + 1j * dH_vii[1]))
    return out


# from gpaw.utilities.debug import frozen
# @frozen
class Hamiltonian:

    def __init__(self, gd, finegd, nspins, collinear, setups, timer, xc, world,
                 redistributor, vext=None):
        self.gd = gd
        self.finegd = finegd
        self.nspins = nspins
        self.collinear = collinear
        self.setups = setups
        self.timer = timer
        self.xc = xc
        self.world = world
        self.redistributor = redistributor

        self.ncomponents = self.nspins if self.collinear else 1 + 3

        self.atomdist = None
        self.dH_asp = None
        self.vt_xG = None
        self.vt_sG = None
        self.vt_vG = None
        self.vHt_g = None
        self.vt_xg = None
        self.vt_sg = None
        self.vt_vg = None
        self.atom_partition = None

        # Energy contributioons that sum up to e_total_free:
        self.e_kinetic = None
        self.e_coulomb = None
        self.e_zero = None
        self.e_external = None
        self.e_xc = None
        self.e_entropy = None
        self.e_band = None

        self.e_total_free = None
        self.e_total_extrapolated = None
        self.e_kinetic0 = None

        self.ref_vt_sG = None
        self.ref_dH_asp = None

        if isinstance(vext, dict):
            vext = create_external_potential(**vext)
        self.vext = vext  # external potential

        self.positions_set = False
        self.spos_ac = None
        self.soc = False

    @property
    def dH(self):
        return functools.partial(apply_non_local_hamilton,
                                 self.dH_asp, self.collinear)

    def update_atomic_hamiltonians(self, value):
        if isinstance(value, dict):
            dtype = complex if self.soc else float
            tmp = self.setups.empty_atomic_matrix(self.ncomponents,
                                                  self.atom_partition, dtype)
            tmp.update(value)
            value = tmp
        assert isinstance(value, ArrayDict) or value is None, type(value)
        if value is not None:
            value.check_consistency()
        self.dH_asp = value

    def __str__(self):
        s = 'Hamiltonian:\n'
        s += ('  XC and Coulomb potentials evaluated on a {0}*{1}*{2} grid\n'
              .format(*self.finegd.N_c))
        s += '  Using the %s Exchange-Correlation functional\n' % self.xc.name
        # We would get the description of the XC functional here,
        # except the thing has probably not been fully initialized yet.
        if self.vext is not None:
            s += '  External potential:\n    {0}\n'.format(self.vext)
        return s

    def summary(self, wfs, log):
        log('Energy contributions relative to reference atoms:',
            '(reference = {0:.6f})\n'.format(self.setups.Eref * Ha))

        energies = [('Kinetic:      ', self.e_kinetic),
                    ('Potential:    ', self.e_coulomb),
                    ('External:     ', self.e_external),
                    ('XC:           ', self.e_xc),
                    ('Entropy (-ST):', self.e_entropy),
                    ('Local:        ', self.e_zero)]

        for name, e in energies:
            log('%-14s %+11.6f' % (name, Ha * e))

        log('--------------------------')
        log('Free energy:   %+11.6f' % (Ha * self.e_total_free))
        log('Extrapolated:  %+11.6f' % (Ha * self.e_total_extrapolated))
        log()
        self.xc.summary(log)

        try:
            workfunctions = self.get_workfunctions(wfs)
        except ValueError:
            pass
        else:
            log('Dipole-layer corrected work functions: {:.6f}, {:.6f} eV'
                .format(*np.array(workfunctions) * Ha))
            log()

    def get_workfunctions(self, wfs):
        """
        Returns the work functions, in Hartree, for a dipole-corrected
        simulation. Returns None if no dipole correction is present.
        (wfs can be obtained from calc.wfs)
        """
        try:
            dipole_correction = self.poisson.correction
        except AttributeError:
            raise ValueError(
                'Work function not defined if no field-free region. Consider '
                'using a dipole correction if you are looking for a '
                'work function.')
        c = self.poisson.c  # index of axis perpendicular to dipole-layer
        if not self.gd.pbc_c[c]:
            # zero boundary conditions
            vacuum = 0.0
        else:
            v_q = self.pd3.gather(self.vHt_q)
            if self.pd3.comm.rank == 0:
                axes = (c, (c + 1) % 3, (c + 2) % 3)
                v_g = self.pd3.ifft(v_q, local=True).transpose(axes)
                vacuum = v_g[0].mean()
            else:
                vacuum = np.nan

        fermilevel = wfs.fermi_level
        wf1 = vacuum - fermilevel + dipole_correction
        wf2 = vacuum - fermilevel - dipole_correction
        return np.array([wf1, wf2])

    def set_positions_without_ruining_everything(self, spos_ac,
                                                 atom_partition):
        self.spos_ac = spos_ac
        rank_a = atom_partition.rank_a

        # If both old and new atomic ranks are present, start a blank dict if
        # it previously didn't exist but it will needed for the new atoms.
        # XXX what purpose does this serve?  In what case does it happen?
        # How would one even go about figuring it out?  Why does it all have
        # to be so unreadable? -Ask
        #
        if (self.atom_partition is not None and
            self.dH_asp is None and (rank_a == self.gd.comm.rank).any()):
            self.update_atomic_hamiltonians({})

        if self.atom_partition is not None and self.dH_asp is not None:
            self.timer.start('Redistribute')
            self.dH_asp.redistribute(atom_partition)
            self.timer.stop('Redistribute')

        self.atom_partition = atom_partition
        self.atomdist = self.redistributor.get_atom_distributions(spos_ac)

    def set_positions(self, spos_ac, atom_partition):
        self.vbar.set_positions(spos_ac, atom_partition)
        self.xc.set_positions(spos_ac)
        self.set_positions_without_ruining_everything(spos_ac, atom_partition)
        self.positions_set = True

    def initialize(self):
        self.vt_xg = self.finegd.empty(self.ncomponents)
        self.vt_sg = self.vt_xg[:self.nspins]
        self.vt_vg = self.vt_xg[self.nspins:]
        self.vHt_g = self.finegd.zeros()
        self.vt_xG = self.gd.empty(self.ncomponents)
        self.vt_sG = self.vt_xG[:self.nspins]
        self.vt_vG = self.vt_xG[self.nspins:]

    def update(self, density, wfs=None, kin_en_using_band=True):
        """Calculate effective potential.

        The XC-potential and the Hartree potential are evaluated on
        the fine grid, and the sum is then restricted to the coarse
        grid."""

        self.timer.start('Hamiltonian')
        if self.vt_sg is None:
            with self.timer('Initialize Hamiltonian'):
                self.initialize()

        finegrid_energies = self.update_pseudo_potential(density)
        coarsegrid_e_kinetic = self.calculate_kinetic_energy(density)

        with self.timer('Calculate atomic Hamiltonians'):
            W_aL = self.calculate_atomic_hamiltonians(density)

        atomic_energies = self.update_corrections(density, W_aL)

        # Make energy contributions summable over world:
        finegrid_energies *= self.finegd.comm.size / self.world.size
        coarsegrid_e_kinetic *= self.gd.comm.size / self.world.size
        # (careful with array orderings/contents)

        if 0:
            print('kinetic', atomic_energies[0], coarsegrid_e_kinetic)
            print('coulomb', atomic_energies[1], finegrid_energies[0])
            print('zero', atomic_energies[2], finegrid_energies[1])
            print('xc', atomic_energies[4], finegrid_energies[3])
            print('external', atomic_energies[3], finegrid_energies[2])

        energies = atomic_energies  # kinetic, coulomb, zero, external, xc
        energies[1:] += finegrid_energies  # coulomb, zero, external, xc
        energies[0] += coarsegrid_e_kinetic  # kinetic

        with self.timer('Communicate'):  # time possible load imbalance
            self.world.sum(energies)
        if not kin_en_using_band:
            assert wfs is not None
            with self.timer('New Kinetic Energy'):
                energies[0] = \
                    self.calculate_kinetic_energy_directly(density,
                                                           wfs)

        (self.e_kinetic0, self.e_coulomb, self.e_zero,
         self.e_external, self.e_xc) = energies

        self.timer.stop('Hamiltonian')

    def update_corrections(self, dens, W_aL):
        self.timer.start('Atomic')
        self.update_atomic_hamiltonians(None)  # XXXX

        e_kinetic = 0.0
        e_coulomb = 0.0
        e_zero = 0.0
        e_external = 0.0
        e_xc = 0.0

        D_asp = self.atomdist.to_work(dens.D_asp)
        dtype = complex if self.soc else float
        dH_asp = self.setups.empty_atomic_matrix(self.ncomponents,
                                                 D_asp.partition, dtype)

        for a, D_sp in D_asp.items():
            W_L = W_aL[a]
            setup = self.setups[a]

            if self.nspins == 2:
                D_p = D_sp.sum(0)
            else:
                D_p = D_sp[0]
            dH_p = (setup.K_p + setup.M_p +
                    setup.MB_p + 2.0 * np.dot(setup.M_pp, D_p) +
                    np.dot(setup.Delta_pL, W_L))
            e_kinetic += np.dot(setup.K_p, D_p) + setup.Kc
            e_zero += setup.MB + np.dot(setup.MB_p, D_p)
            e_coulomb += setup.M + np.dot(D_p, (setup.M_p +
                                                np.dot(setup.M_pp, D_p)))

            if self.soc:
                dH_vii = soc(setup, self.xc, D_sp)
                dH_sp = np.zeros_like(D_sp, dtype=complex)
                dH_sp[1:] = pack2(dH_vii)
            else:
                dH_sp = np.zeros_like(D_sp)

            if setup.hubbard_u is not None:
                eU, dHU_sp = setup.hubbard_u.calculate(setup, D_sp)
                e_xc += eU
                dH_sp += dHU_sp

            dH_sp[:self.nspins] += dH_p

            if self.vext is not None:
                self.vext.paw_correction(setup.Delta_pL[:, 0], dH_sp)

            if self.vext and self.vext.get_name() == 'CDFTPotential':
                # cDFT atomic hamiltonian, eq. 25
                # energy correction added in cDFT main
                h_cdft_a, h_cdft_b = self.vext.get_atomic_hamiltonians(
                    setups=setup.Delta_pL[:, 0], atom=a)

                dH_sp[0] += h_cdft_a
                dH_sp[1] += h_cdft_b

            if self.ref_dH_asp:
                assert self.collinear
                dH_sp += self.ref_dH_asp[a]

            dH_asp[a] = dH_sp

        self.timer.start('XC Correction')
        for a, D_sp in D_asp.items():
            e_xc += self.xc.calculate_paw_correction(self.setups[a], D_sp,
                                                     dH_asp[a], a=a)
        self.timer.stop('XC Correction')
        for a, D_sp in D_asp.items():
            e_kinetic -= (D_sp * dH_asp[a]).sum().real

        self.update_atomic_hamiltonians(self.atomdist.from_work(dH_asp))
        self.timer.stop('Atomic')

        # Make corrections due to non-local xc:
        # self.Enlxc = 0.0  # XXXxcfunc.get_non_local_energy()
        e_kinetic += self.xc.get_kinetic_energy_correction() / self.world.size

        return np.array([e_kinetic, e_coulomb, e_zero, e_external, e_xc])

    def get_energy(self, e_entropy, wfs, kin_en_using_band=True):
        """Sum up all eigenvalues weighted with occupation numbers"""
        self.e_band = wfs.calculate_band_energy()
        if kin_en_using_band:
            self.e_kinetic = self.e_kinetic0 + self.e_band
        else:
            self.e_kinetic = self.e_kinetic0
        self.e_entropy = e_entropy
        if 0:
            print(self.e_kinetic0,
                  self.e_band,
                  self.e_coulomb,
                  self.e_external,
                  self.e_zero,
                  self.e_xc,
                  self.e_entropy)

        self.e_total_free = (self.e_kinetic + self.e_coulomb +
                             self.e_external + self.e_zero + self.e_xc +
                             self.e_entropy)
        self.e_total_extrapolated = (
            self.e_total_free +
            wfs.occupations.extrapolate_factor * e_entropy)

        return self.e_total_free

    def linearize_to_xc(self, new_xc, density):
        # Store old hamiltonian
        ref_vt_sG = self.vt_sG.copy()
        ref_dH_asp = self.dH_asp.copy()
        self.xc = new_xc
        self.xc.set_positions(self.spos_ac)
        self.update(density)

        ref_vt_sG -= self.vt_sG
        for a, dH_sp in self.dH_asp.items():
            ref_dH_asp[a] -= dH_sp
        self.ref_vt_sG = ref_vt_sG
        self.ref_dH_asp = self.atomdist.to_work(ref_dH_asp)

    def calculate_forces(self, dens, F_av):
        ghat_aLv = dens.ghat.dict(derivative=True)
        nct_av = dens.nct.dict(derivative=True)
        vbar_av = self.vbar.dict(derivative=True)

        self.calculate_forces2(dens, ghat_aLv, nct_av, vbar_av)
        F_coarsegrid_av = np.zeros_like(F_av)

        # Force from compensation charges:
        _Q, Q_aL = dens.calculate_multipole_moments()
        for a, dF_Lv in ghat_aLv.items():
            F_av[a] += np.dot(Q_aL[a], dF_Lv)

        # Force from smooth core charge:
        for a, dF_v in nct_av.items():
            F_coarsegrid_av[a] += dF_v[0]

        # Force from zero potential:
        for a, dF_v in vbar_av.items():
            F_av[a] += dF_v[0]

        self.xc.add_forces(F_av)
        self.gd.comm.sum(F_coarsegrid_av, 0)
        self.finegd.comm.sum(F_av, 0)
        if self.vext:
            if self.vext.get_name() == 'CDFTPotential':
                F_av += self.vext.get_cdft_forces()
        F_av += F_coarsegrid_av

    def apply_local_potential(self, psit_nG, Htpsit_nG, s):
        vt_G = self.vt_sG[s]
        if psit_nG.ndim == 3:
            Htpsit_nG += psit_nG * vt_G
        else:
            for psit_G, Htpsit_G in zip(psit_nG, Htpsit_nG):
                Htpsit_G += psit_G * vt_G

    def apply(self, a_xG, b_xG, wfs, kpt, calculate_P_ani=True):
        """Apply the Hamiltonian operator to a set of vectors.

        Parameters:

        a_nG: ndarray
            Set of vectors to which the overlap operator is applied.
        b_nG: ndarray, output
            Resulting S times a_nG vectors.
        wfs: WaveFunctions
            Wave-function object defined in wavefunctions.py
        kpt: KPoint object
            k-point object defined in kpoint.py.
        calculate_P_ani: bool
            When True, the integrals of projector times vectors
            P_ni = <p_i | a_nG> are calculated.
            When False, existing P_ani are used

        """

        wfs.kin.apply(a_xG, b_xG, kpt.phase_cd)
        self.apply_local_potential(a_xG, b_xG, kpt.s)
        shape = a_xG.shape[:-3]
        P_axi = wfs.pt.dict(shape)

        if calculate_P_ani:  # TODO calculate_P_ani=False is experimental
            wfs.pt.integrate(a_xG, P_axi, kpt.q)
        else:
            for a, P_ni in kpt.P_ani.items():
                P_axi[a][:] = P_ni

        for a, P_xi in P_axi.items():
            dH_ii = unpack(self.dH_asp[a][kpt.s])
            P_axi[a] = np.dot(P_xi, dH_ii)
        wfs.pt.add(b_xG, P_axi, kpt.q)

    def get_xc_difference(self, xc, density):
        """Calculate non-selfconsistent XC-energy difference."""
        if density.nt_sg is None:
            density.interpolate_pseudo_density()
        nt_sg = density.nt_sg
        if hasattr(xc, 'hybrid'):
            xc.calculate_exx()
        finegd_e_xc = xc.calculate(density.finegd, nt_sg)
        D_asp = self.atomdist.to_work(density.D_asp)
        atomic_e_xc = 0.0
        for a, D_sp in D_asp.items():
            setup = self.setups[a]
            atomic_e_xc += xc.calculate_paw_correction(setup, D_sp, a=a)
        e_xc = finegd_e_xc + self.world.sum(atomic_e_xc)
        return e_xc - self.e_xc

    def estimate_memory(self, mem):
        nbytes = self.gd.bytecount()
        nfinebytes = self.finegd.bytecount()
        arrays = mem.subnode('Arrays', 0)
        arrays.subnode('vHt_g', nfinebytes)
        arrays.subnode('vt_sG', self.nspins * nbytes)
        arrays.subnode('vt_sg', self.nspins * nfinebytes)
        self.xc.estimate_memory(mem.subnode('XC'))
        self.poisson.estimate_memory(mem.subnode('Poisson'))
        self.vbar.estimate_memory(mem.subnode('vbar'))

    def write(self, writer):
        # Write all eneriges:
        for name in ENERGY_NAMES:
            energy = getattr(self, name)
            if energy is not None:
                energy *= Ha
            writer.write(name, energy)

        writer.write(
            potential=self.gd.collect(self.vt_xG) * Ha,
            atomic_hamiltonian_matrices=pack_atomic_matrices(self.dH_asp) * Ha)

        self.xc.write(writer.child('xc'))

        if hasattr(self.poisson, 'write'):
            self.poisson.write(writer.child('poisson'))

    def read(self, reader):
        h = reader.hamiltonian

        # Read all energies:
        for name in ENERGY_NAMES:
            energy = h.get(name)
            if energy is not None:
                energy /= reader.ha
            setattr(self, name, energy)

        # Read pseudo potential on the coarse grid
        # and broadcast on kpt/band comm:
        self.initialize()
        self.gd.distribute(h.potential / reader.ha, self.vt_xG)

        self.atom_partition = AtomPartition(self.gd.comm,
                                            np.zeros(len(self.setups), int),
                                            name='hamiltonian-init-serial')

        # Read non-local part of hamiltonian
        self.update_atomic_hamiltonians({})
        dH_sP = h.atomic_hamiltonian_matrices / reader.ha

        if self.gd.comm.rank == 0:
            self.update_atomic_hamiltonians(
                unpack_atomic_matrices(dH_sP, self.setups))

        if hasattr(self.poisson, 'read'):
            self.poisson.read(reader)
            self.poisson.set_grid_descriptor(self.finegd)

    def calculate_kinetic_energy_directly(self, density, wfs):

        """
        Calculate kinetic energy as 1/2 (nable psi)^2
        it gives better estimate of kinetic energy during the SCF.
        Important for direct min.

        'calculate_kinetic_energy' method gives a correct
        value of kinetic energy only at self-consistent solution.

        :param density:
        :param wfs:
        :return: total kinetic energy
        """
        # pseudo-part
        if wfs.mode == 'lcao':
            return self.calculate_kinetic_energy_using_kin_en_matrix(
                density, wfs)
        elif wfs.mode == 'pw':
            e_kin = 0.0
            for kpt in wfs.kpt_u:
                for f, psit_G in zip(kpt.f_n, kpt.psit_nG):
                    if f > 1.0e-10:
                        G2_G = wfs.pd.G2_qG[kpt.q]
                        e_kin += f * wfs.pd.integrate(
                            0.5 * G2_G * psit_G, psit_G).real
        else:
            e_kin = 0.0

            def Lapl(psit_G, kpt):
                Lpsit_G = np.zeros_like(psit_G)
                wfs.kin.apply(psit_G, Lpsit_G, kpt.phase_cd)
                return Lpsit_G

            for kpt in wfs.kpt_u:
                for f, psit_G in zip(kpt.f_n, kpt.psit_nG):
                    if f > 1.0e-10:
                        e_kin += f * wfs.integrate(
                            Lapl(psit_G, kpt), psit_G, False)
            e_kin = e_kin.real
            e_kin = wfs.gd.comm.sum(e_kin)

        e_kin = wfs.kd.comm.sum(e_kin)  # ?
        # paw corrections
        e_kin_paw = 0.0
        for a, D_sp in density.D_asp.items():
            setup = wfs.setups[a]
            D_p = D_sp.sum(0)
            e_kin_paw += np.dot(setup.K_p, D_p) + setup.Kc
        e_kin_paw = density.gd.comm.sum(e_kin_paw)
        return e_kin + e_kin_paw

    def calculate_kinetic_energy_using_kin_en_matrix(self, density,
                                                     wfs):
        """
        E_k = sum_{M'M} rho_MM' T_M'M
        better agreement between gradients of energy and
        the total energy during the direct minimisation.
        This is important when the line search is used.
        Also avoids using the eigenvalues which are
        not calculated during the direct minimisation.

        'calculate_kinetic_energy' method gives a correct
        value of kinetic energy only at self-consistent solution.

        :param density:
        :param wfs:
        :return: total kinetic energy
        """
        # pseudo-part
        e_kinetic = 0.0
        e_kin_paw = 0.0

        for kpt in wfs.kpt_u:
            # calculation of the density matrix directly
            # can be expansive for a large scale
            # as there are lot of empty states
            # when the exponential transformation is used
            # (n_bands=n_basis_functions.)
            #
            # rho_MM = \
            #     wfs.calculate_density_matrix(kpt.f_n, kpt.C_nM)
            # e_kinetic += np.einsum('ij,ji->', kpt.T_MM, rho_MM)
            #
            # the code below is faster
            self.timer.start('Pseudo part')
            occ = kpt.f_n > 1e-10
            x_nn = np.dot(kpt.C_nM[occ],
                          np.dot(kpt.T_MM,
                                 kpt.C_nM[occ].T.conj())).real
            e_kinetic += np.einsum('i,ii->', kpt.f_n[occ], x_nn)
            self.timer.stop('Pseudo part')
        # del rho_MM

        e_kinetic = wfs.kd.comm.sum(e_kinetic)
        # paw corrections
        for a, D_sp in density.D_asp.items():
            setup = wfs.setups[a]
            D_p = D_sp.sum(0)
            e_kin_paw += np.dot(setup.K_p, D_p) + setup.Kc

        e_kin_paw = self.gd.comm.sum(e_kin_paw)

        return e_kinetic.real + e_kin_paw


class RealSpaceHamiltonian(Hamiltonian):
    def __init__(self, gd, finegd, nspins, collinear, setups, timer, xc, world,
                 vext=None,
                 psolver=None, stencil=3, redistributor=None,
                 charge: float = 0.0):
        Hamiltonian.__init__(self, gd, finegd, nspins, collinear,
                             setups, timer, xc,
                             world, vext=vext,
                             redistributor=redistributor)

        # Solver for the Poisson equation:
        if psolver is None:
            psolver = {}
        if isinstance(psolver, dict):
            psolver = PoissonSolver(**psolver)
        self.poisson = psolver
        self.poisson.set_grid_descriptor(self.finegd)

        # Restrictor function for the potential:
        self.restrictor = Transformer(self.finegd, self.redistributor.aux_gd,
                                      stencil)
        self.restrict = self.restrictor.apply

        self.vbar = LFC(self.finegd, [[setup.vbar] for setup in setups],
                        forces=True)

        self.vbar_g = None
        self.npoisson = None

    def restrict_and_collect(self, a_xg, b_xg=None, phases=None):
        if self.redistributor.enabled:
            tmp_xg = self.restrictor.apply(a_xg, output=None, phases=phases)
            b_xg = self.redistributor.collect(tmp_xg, b_xg)
        else:
            b_xg = self.restrictor.apply(a_xg, output=b_xg, phases=phases)
        return b_xg

    def __str__(self):
        s = Hamiltonian.__str__(self)

        degree = self.restrictor.nn * 2 - 1
        name = ['linear', 'cubic', 'quintic', 'heptic'][degree // 2]
        s += ('  Interpolation: tri-%s ' % name +
              '(%d. degree polynomial)\n' % degree)
        s += '  Poisson solver: %s' % self.poisson.get_description()
        return s

    def set_positions(self, spos_ac, rank_a):
        Hamiltonian.set_positions(self, spos_ac, rank_a)
        if self.vbar_g is None:
            self.vbar_g = self.finegd.empty()
        self.vbar_g[:] = 0.0
        self.vbar.add(self.vbar_g)

    def update_pseudo_potential(self, dens):
        self.timer.start('vbar')
        e_zero = self.finegd.integrate(self.vbar_g, dens.nt_g,
                                       global_integral=False)

        vt_g = self.vt_sg[0]
        vt_g[:] = self.vbar_g
        self.timer.stop('vbar')

        e_external = 0.0
        if self.vext is not None:
            if self.vext.get_name() == 'CDFTPotential':
                vext_g = self.vext.get_potential(self.finegd).copy()
                e_external += self.vext.get_cdft_external_energy(
                    dens,
                    self.nspins, vext_g, vt_g, self.vbar_g, self.vt_sg)

            else:
                vext_g = self.vext.get_potential(self.finegd)
                vt_g += vext_g
                e_external = self.finegd.integrate(vext_g, dens.rhot_g,
                                                   global_integral=False)

        if self.nspins == 2:
            self.vt_sg[1] = vt_g

        self.timer.start('XC 3D grid')
        e_xc = self.xc.calculate(self.finegd, dens.nt_sg, self.vt_sg)
        e_xc /= self.finegd.comm.size
        self.timer.stop('XC 3D grid')

        self.timer.start('Poisson')
        # npoisson is the number of iterations:
        self.npoisson = self.poisson.solve(self.vHt_g, dens.rhot_g,
                                           charge=-dens.charge,
                                           timer=self.timer)
        self.timer.stop('Poisson')

        self.timer.start('Hartree integrate/restrict')

        e_coulomb = 0.5 * self.finegd.integrate(self.vHt_g, dens.rhot_g,
                                                global_integral=False)

        for vt_g in self.vt_sg:
            vt_g += self.vHt_g

        self.timer.stop('Hartree integrate/restrict')
        energies = np.array([e_coulomb, e_zero, e_external, e_xc])
        return energies

    def calculate_kinetic_energy(self, density):
        # XXX new timer item for kinetic energy?
        self.timer.start('Hartree integrate/restrict')
        self.restrict_and_collect(self.vt_sg, self.vt_sG)

        e_kinetic = 0.0
        s = 0
        for vt_G, nt_G in zip(self.vt_sG, density.nt_sG):
            if self.ref_vt_sG is not None:
                vt_G += self.ref_vt_sG[s]

            if s < self.nspins:
                e_kinetic -= self.gd.integrate(vt_G, nt_G - density.nct_G,
                                               global_integral=False)
            else:
                e_kinetic -= self.gd.integrate(vt_G, nt_G,
                                               global_integral=False)
            s += 1
        self.timer.stop('Hartree integrate/restrict')
        return e_kinetic

    def calculate_atomic_hamiltonians(self, dens):
        def getshape(a):
            return sum(2 * l + 1 for l, _ in enumerate(self.setups[a].ghat_l)),
        W_aL = ArrayDict(self.atomdist.aux_partition, getshape, float)
        if self.vext:
            if self.vext.get_name() != 'CDFTPotential':
                vext_g = self.vext.get_potential(self.finegd)
                dens.ghat.integrate(self.vHt_g + vext_g, W_aL)
            else:
                dens.ghat.integrate(self.vHt_g, W_aL)
        else:
            dens.ghat.integrate(self.vHt_g, W_aL)

        return self.atomdist.to_work(self.atomdist.from_aux(W_aL))

    def calculate_forces2(self, dens, ghat_aLv, nct_av, vbar_av):
        if self.nspins == 2:
            vt_G = self.vt_sG.mean(0)
        else:
            vt_G = self.vt_sG[0]

        self.vbar.derivative(dens.nt_g, vbar_av)
        if self.vext:
            if self.vext.get_name() == 'CDFTPotential':
                # CDFT force added in calculate_forces
                dens.ghat.derivative(self.vHt_g, ghat_aLv)
            else:
                vext_g = self.vext.get_potential(self.finegd)
                dens.ghat.derivative(self.vHt_g + vext_g, ghat_aLv)
        else:
            dens.ghat.derivative(self.vHt_g, ghat_aLv)
        dens.nct.derivative(vt_G, nct_av)

    def get_electrostatic_potential(self, dens):
        self.update(dens)

        v_g = self.finegd.collect(self.vHt_g, broadcast=True)
        v_g = self.finegd.zero_pad(v_g)
        if hasattr(self.poisson, 'correction'):
            assert self.poisson.c == 2
            v_g[:, :, 0] = self.poisson.correction
        return v_g



from __future__ import annotations
from abc import ABC, abstractmethod

import numpy as np

from gpaw.sphere.integrate import spherical_truncation_function_collection
from gpaw.sphere.rshe import calculate_reduced_rshe

from gpaw.response import timer
from gpaw.response.kspair import KohnShamKPointPair
from gpaw.response.pair import phase_shifted_fft_indices
from gpaw.response.site_paw import calculate_site_matrix_element_correction
from gpaw.response.localft import extract_micro_setup, calculate_LSDA_Wxc


class MatrixElement(ABC):
    """Data class for transitions distributed Kohn-Sham matrix elements."""

    def __init__(self, tblocks, qpd):
        self.tblocks = tblocks
        self.qpd = qpd

        self.array = self.zeros()
        assert self.array.shape[0] == tblocks.blocksize

    @abstractmethod
    def zeros(self):
        """Generate matrix element array with zeros."""

    @property
    def local_array_view(self):
        return self.array[:self.tblocks.nlocal]

    def get_global_array(self):
        """Get the global (all gathered) matrix element."""
        return self.tblocks.all_gather(self.array)


class MatrixElementCalculator(ABC):
    r"""Abstract base class for matrix element calculators.

    In the PAW method, Kohn-Sham matrix elements,
                            ˰
    A_(nks,n'k's') = <ψ_nks|A|ψ_n'k's'>

    can be evaluated in the space of pseudo waves using the pseudo operator
            __  __
    ˷   ˰   \   \   ˷           ˰           ˷    ˰ ˷       ˷
    A = A + /   /  |p_ai>[<φ_ai|A|φ_ai'> - <φ_ai|A|φ_ai'>]<p_ai'|
            ‾‾  ‾‾
            a   i,i'

    to which effect,
                      ˷     ˷ ˷
    A_(nks,n'k's') = <ψ_nks|A|ψ_n'k's'>

    This is an abstract base class for calculating such matrix elements for a
    number of band and spin transitions t=(n,s)->(n',s') for a given k-point
    pair k and k + q:

    A_kt = A_(nks,n'k+qs')
    """

    def __init__(self, gs, context):
        self.gs = gs
        self.context = context

    @timer('Calculate matrix element')
    def __call__(self, kptpair: KohnShamKPointPair, qpd) -> MatrixElement:
        r"""Calculate the matrix element for all transitions t.

        The calculation is split into a pseudo contribution and a PAW
        correction:
               ˷
        A_kt = A_kt + ΔA_kt,

        see [PRB 103, 245110 (2021)] for additional details and references.
        """
        matrix_element = self.create_matrix_element(kptpair.tblocks, qpd)
        self.add_pseudo_contribution(kptpair, matrix_element)
        self.add_paw_correction(kptpair, matrix_element)

        return matrix_element

    @abstractmethod
    def create_matrix_element(self, tblocks, qpd):
        """Return a new MatrixElement instance."""

    def add_pseudo_contribution(self, kptpair, matrix_element):
        """Add the pseudo matrix element to an output array.

        The pseudo matrix element is evaluated on the coarse real-space grid
        and integrated:

        ˷       ˷     ˰ ˷
        A_kt = <ψ_nks|A|ψ_n'k+qs'>

               /    ˷          ˰ ˷
             = | dr ψ_nks^*(r) A ψ_n'k+qs'(r)
               /
        """
        ikpt1 = kptpair.ikpt1
        ikpt2 = kptpair.ikpt2

        # Map the k-points from the irreducible part of the BZ to the BZ
        # k-point K (up to a reciprocal lattice vector)
        k1_c = self.gs.ibz2bz[kptpair.K1].map_kpoint()
        k2_c = self.gs.ibz2bz[kptpair.K2].map_kpoint()

        # Fourier transform the periodic part of the pseudo waves to the coarse
        # real-space grid and map them to the BZ k-point K (up to the same
        # reciprocal lattice vector as above)
        ut1_hR = self.get_periodic_pseudo_waves(kptpair.K1, ikpt1)
        ut2_hR = self.get_periodic_pseudo_waves(kptpair.K2, ikpt2)

        # Fold out the pseudo waves to the transition index
        ut1_mytR = ut1_hR[ikpt1.h_myt]
        ut2_mytR = ut2_hR[ikpt2.h_myt]

        self._add_pseudo_contribution(
            k1_c, k2_c, ut1_mytR, ut2_mytR, matrix_element)

    def add_paw_correction(self, kptpair, matrix_element):
        r"""Add the matrix element PAW correction to an output array.

        The PAW correction is calculated using the projector overlaps of the
        pseudo waves:
                __  __
                \   \   ˷     ˷              ˷     ˷
        ΔA_kt = /   /  <ψ_nks|p_ai> ΔA_aii' <p_ai'|ψ_n'k+qs'>
                ‾‾  ‾‾
                a   i,i'

        where the PAW correction tensor is calculated on a radial grid inside
        each augmentation sphere of position R_a, using the atom-centered
        partial waves φ_ai(r):
                        ˰           ˷    ˰ ˷
        ΔA_aii' = <φ_ai|A|φ_ai'> - <φ_ai|A|φ_ai'>

                  /                   ˰
                = | dr [φ_ai^*(r-R_a) A φ_ai'(r-R_a)
                  /       ˷             ˰ ˷
                        - φ_ai^*(r-R_a) A φ_ai'(r-R_a)]
        """
        ikpt1 = kptpair.ikpt1
        ikpt2 = kptpair.ikpt2

        # Map the projections from the irreducible part of the BZ to the BZ
        # k-point K
        P1h = self.gs.ibz2bz[kptpair.K1].map_projections(ikpt1.Ph)
        P2h = self.gs.ibz2bz[kptpair.K2].map_projections(ikpt2.Ph)

        # Fold out the projectors to the transition index
        P1_amyti = ikpt1.projectors_in_transition_index(P1h)
        P2_amyti = ikpt2.projectors_in_transition_index(P2h)
        assert P1_amyti.atom_partition.comm.size == \
            P2_amyti.atom_partition.comm.size == 1, \
            'We need access to the projections of all atoms'

        self._add_paw_correction(P1_amyti, P2_amyti, matrix_element)

    @abstractmethod
    def _add_pseudo_contribution(self, k1_c, k2_c, ut1_mytR, ut2_mytR,
                                 matrix_element):
        """Add pseudo contribution based on the pseudo waves in real space."""

    @abstractmethod
    def _add_paw_correction(self, P1_amyti, P2_amyti, matrix_element):
        """Add paw correction based on the projector overlaps."""

    def get_periodic_pseudo_waves(self, K, ikpt):
        """FFT the Kohn-Sham orbitals to real space and map them from the
        irreducible k-point to the k-point in question."""
        ut_hR = self.gs.gd.empty(ikpt.nh, self.gs.dtype)
        for h, psit_G in enumerate(ikpt.psit_hG):
            ut_hR[h] = self.gs.ibz2bz[K].map_pseudo_wave(
                self.gs.global_pd.ifft(psit_G, ikpt.ik))

        return ut_hR


class PairDensity(MatrixElement):

    def zeros(self):
        return self.qpd.zeros(self.tblocks.blocksize)


class NewPairDensityCalculator(MatrixElementCalculator):
    r"""Class for calculating pair densities

    n_kt(G+q) = n_nks,n'k+qs'(G+q) = <nks| e^-i(G+q)r |n'k+qs'>

                /
              = | dr e^-i(G+q)r ψ_nks^*(r) ψ_n'k+qs'(r)
                /

    for a single k-point pair (k, k + q) in the plane-wave mode.

    As always, the calculation is split in a pseudo contribution and a PAW
    correction, n_kt(G+q) = ñ_kt(G+q) + Δn_kt(G+q).
    """

    def __init__(self, gs, context):
        super().__init__(gs, context)

        # Save PAW correction for all calls with same q_c
        self._pawcorr = None
        self._currentq_c = None

    def initialize_paw_corrections(self, qpd):
        """Initialize the PAW corrections ahead of the actual calculation."""
        self.get_paw_corrections(qpd)

    def get_paw_corrections(self, qpd):
        """Get PAW corrections correcsponding to a specific q-vector."""
        if self._pawcorr is None \
           or not np.allclose(qpd.q_c - self._currentq_c, 0.):
            with self.context.timer('Initialize PAW corrections'):
                self._pawcorr = self.gs.pair_density_paw_corrections(qpd)
                self._currentq_c = qpd.q_c

        return self._pawcorr

    @staticmethod
    def create_matrix_element(tblocks, qpd):
        return PairDensity(tblocks, qpd)

    @timer('Calculate the pseudo pair density')
    def _add_pseudo_contribution(self, k1_c, k2_c, ut1_mytR, ut2_mytR,
                                 pair_density):
        r"""Add the pseudo pair density to an output array.

        The pseudo pair density is first evaluated on the coarse real-space
        grid and then FFT'ed to reciprocal space,

                    /               ˷          ˷
        ñ_kt(G+q) = | dr e^-i(G+q)r ψ_nks^*(r) ψ_n'k+qs'(r)
                    /V0
                                 ˷          ˷
                  = FFT_G[e^-iqr ψ_nks^*(r) ψ_n'k+qs'(r)]

        where the Kohn-Sham orbitals are normalized to the unit cell.
        """
        qpd = pair_density.qpd
        n_mytG = pair_density.local_array_view

        # Calculate the pseudo pair density in real space, up to a phase of
        # e^(-i[k+q-k']r).
        # This phase does not necessarily vanish, since k2_c only is required
        # to equal k1_c + qpd.q_c modulo a reciprocal lattice vector.
        nt_mytR = ut1_mytR.conj() * ut2_mytR

        # Get the FFT indices corresponding to the Fourier transform
        #                       ˷          ˷
        # FFT_G[e^(-i[k+q-k']r) u_nks^*(r) u_n'k's'(r)]
        Q_G = phase_shifted_fft_indices(k1_c, k2_c, qpd)

        # Add the desired plane-wave components of the FFT'ed pseudo pair
        # density to the output array
        for n_G, n_R in zip(n_mytG, nt_mytR):
            n_G[:] += qpd.fft(n_R, 0, Q_G) * qpd.gd.dv

    @timer('Calculate the pair density PAW corrections')
    def _add_paw_correction(self, P1_amyti, P2_amyti, pair_density):
        r"""Add the pair-density PAW correction to the output array.

        The correction is calculated from
                     __  __
                     \   \   ˷     ˷     ˷    ˷
        Δn_kt(G+q) = /   /  <ψ_nks|p_ai><p_ai'|ψ_n'k+qs'> Q_aii'(G+q)
                     ‾‾  ‾‾
                     a   i,i'

        where the pair-density PAW correction tensor is given by:

                      /
        Q_aii'(G+q) = | dr e^-i(G+q)r [φ_ai^*(r-R_a) φ_ai'(r-R_a)
                      /                  ˷             ˷
                                       - φ_ai^*(r-R_a) φ_ai'(r-R_a)]
        """
        n_mytG = pair_density.local_array_view
        Q_aGii = self.get_paw_corrections(pair_density.qpd).Q_aGii
        for a, Q_Gii in enumerate(Q_aGii):
            # Make outer product of the projector overlaps
            P1ccP2_mytii = P1_amyti[a].conj()[..., np.newaxis] \
                * P2_amyti[a][:, np.newaxis]
            # Sum over partial wave indices and add correction to the output
            n_mytG[:] += np.einsum('tij, Gij -> tG', P1ccP2_mytii, Q_Gii)


class SiteMatrixElement(MatrixElement):
    def __init__(self, tblocks, qpd, atomic_site_data):
        self.nsites = atomic_site_data.nsites
        self.npartitions = atomic_site_data.npartitions
        super().__init__(tblocks, qpd)

    def zeros(self):
        return np.zeros(
            (self.tblocks.blocksize, self.nsites, self.npartitions),
            dtype=complex)


class SiteMatrixElementCalculator(MatrixElementCalculator):
    r"""Class for calculating site matrix elements.

    The site matrix elements are defined as the expectation value of any local
    functional of the electron density f[n](r) = f(n(r)), evaluated on a given
    site a for every site partitioning p. The sites are defined in terms of
    smooth truncation functions Θ(r∊Ω_ap), interpolating smoothly between unity
    for positions inside the spherical site volume and zero outside it:

    f^ap_kt = f^ap_(nks,n'k+qs') = <ψ_nks|Θ(r∊Ω_ap)f(r)|ψ_n'k+qs'>

             /
           = | dr Θ(r∊Ω_ap) f(r) ψ_nks^*(r) ψ_n'k+qs'(r)
             /

    For details, see [publication in preparation].
    """

    def __init__(self, gs, context, atomic_site_data,
                 rshelmax: int = -1,
                 rshewmin: float | None = None):
        """Construct the SiteMatrixElementCalculator.

        Parameters
        ----------
        rshelmax : int
            The maximum index l (l < 6) to use in the expansion of f(r) into
            real spherical harmonics for the PAW correction.
        rshewmin : float or None
            If None, the PAW correction will be fully expanded up to the chosen
            lmax. Given as a float (0 < rshewmin < 1), rshewmin indicates what
            coefficients to use in the expansion. If any (l,m) coefficient
            contributes with less than a fraction of rshewmin on average, it
            will not be included.
        """
        super().__init__(gs, context)
        self.atomic_site_data = atomic_site_data

        self.rshelmax = rshelmax
        self.rshewmin = rshewmin

        # PAW correction tensor
        self._F_apii = None

    @abstractmethod
    def add_f(gd, n_sx, f_x):
        """Add the local functional f(n(r)) to the f_x output array."""

    def get_paw_correction_tensor(self):
        if self._F_apii is None:
            self._F_apii = self.calculate_paw_correction_tensor()
        return self._F_apii

    def calculate_paw_correction_tensor(self):
        """Calculate the site matrix element correction tensor F_ii'^ap."""
        F_apii = []
        adata = self.atomic_site_data
        for a, (A, rc_p, lambd_p) in enumerate(zip(
                adata.A_a, adata.rc_ap, adata.lambd_ap)):
            # Expand local function in real spherical harmonics
            micro_setup = extract_micro_setup(self.gs, A)
            rshe, info_string = self.perform_rshe(micro_setup)
            self.print_rshe_info(a, A, info_string)

            # Calculate the PAW correction
            pawdata = self.gs.pawdatasets[A]
            F_apii.append(calculate_site_matrix_element_correction(
                pawdata, rshe, rc_p, adata.drcut, lambd_p))

        return F_apii

    def perform_rshe(self, micro_setup):
        """Expand the functional f(n(r)) into real spherical harmonics."""
        f_ng = micro_setup.evaluate_function(self.add_f)
        return calculate_reduced_rshe(
            micro_setup.rgd, f_ng, micro_setup.Y_nL,
            self.rshelmax, self.rshewmin)

    def print_rshe_info(self, a, A, info_string):
        """Print information about the expansion at site a."""
        info_string = f'RSHE of site {a} (atom {A}):\n' + info_string
        self.context.print(info_string.replace('\n', '\n    ') + '\n')

    def create_matrix_element(self, tblocks, qpd):
        return SiteMatrixElement(tblocks, qpd, self.atomic_site_data)

    @timer('Calculate pseudo site matrix element')
    def _add_pseudo_contribution(self, k1_c, k2_c, ut1_mytR, ut2_mytR,
                                 site_matrix_element):
        """Add the pseudo site matrix element to the output array.

        The pseudo matrix element is evaluated on the coarse real-space grid
        and integrated together with the smooth truncation function,

        ˷         /                   ˷          ˷
        f^ap_kt = | dr Θ(r∊Ω_ap) f(r) ψ_nks^*(r) ψ_n'k+qs'(r)
                  /

        where the Kohn-Sham orbitals are normalized to the unit cell.
        """
        # Construct pseudo waves with Bloch phases
        r_Rc = np.transpose(self.gs.ibz2bz.r_cR,  # scaled grid coordinates
                            (1, 2, 3, 0))
        psit1_mytR = np.exp(2j * np.pi * r_Rc @ k1_c)[np.newaxis] * ut1_mytR
        psit2_mytR = np.exp(2j * np.pi * r_Rc @ k2_c)[np.newaxis] * ut2_mytR
        # Calculate real-space pair densities ñ_kt(r)
        nt_mytR = psit1_mytR.conj() * psit2_mytR
        # Evaluate the local functional f(n(r)) on the coarse real-space grid
        # NB: Here we assume that f(r) is sufficiently smooth to be represented
        # on a regular grid (unlike the wave functions).
        n_sR, gd = self.gs.get_all_electron_density(gridrefinement=1)
        f_R = gd.zeros()
        self.add_f(gd, n_sR, f_R)

        # Set up spherical truncation function collection on the coarse
        # real-space grid with the KPointDescriptor of the q-point.
        adata = self.atomic_site_data
        stfc = spherical_truncation_function_collection(
            self.gs.gd, adata.spos_ac,
            adata.rc_ap, adata.drcut, adata.lambd_ap,
            kd=site_matrix_element.qpd.kd, dtype=complex)

        # Integrate Θ(r∊Ω_ap) f(r) ñ_kt(r)
        ntlocal = nt_mytR.shape[0]
        ft_amytp = {a: np.empty((ntlocal, adata.npartitions), dtype=complex)
                    for a in range(adata.nsites)}
        stfc.integrate(nt_mytR * f_R[np.newaxis], ft_amytp, q=0)

        # Add integral to output array
        f_mytap = site_matrix_element.array
        for a in range(adata.nsites):
            f_mytap[:, a] += ft_amytp[a]

    @timer('Calculate site matrix element PAW correction')
    def _add_paw_correction(self, P1_Amyti, P2_Amyti, site_matrix_element):
        r"""Add the site matrix element PAW correction to the output array.

        For every site a, we only need a PAW correction for that site itself,
                   __
                   \   ˷     ˷              ˷     ˷
        Δf^ap_kt = /  <ψ_nks|p_ai> F_apii' <p_ai'|ψ_n'k+qs'>
                   ‾‾
                   i,i'

        where F_apii' is the site matrix element correction tensor.
        """
        f_mytap = site_matrix_element.array
        F_apii = self.get_paw_correction_tensor()
        for a, (A, F_pii) in enumerate(zip(
                self.atomic_site_data.A_a, F_apii)):
            # Make outer product of the projector overlaps
            P1ccP2_mytii = P1_Amyti[A].conj()[..., np.newaxis] \
                * P2_Amyti[A][:, np.newaxis]
            # Sum over partial wave indices and add correction to the output
            f_mytap[:, a] += np.einsum('tij, pij -> tp', P1ccP2_mytii, F_pii)


class SitePairDensityCalculator(SiteMatrixElementCalculator):
    """Class for calculating site pair densities.

    The site pair density corresponds to a site matrix element with f(r) = 1:

    n^ap_(nks,n'k+qs') = <ψ_nks|Θ(r∊Ω_ap)|ψ_n'k+qs'>
    """

    def __init__(self, gs, context, atomic_site_data):
        super().__init__(gs, context, atomic_site_data,
                         # Expanding f(r) = 1 in real spherical harmonics only
                         # involves l = 0
                         rshelmax=0)

    def add_f(self, gd, n_sx, f_x):
        f_x[:] += 1.

    def print_rshe_info(self, *args):
        # The expansion in spherical harmonics is trivial (l = 0), so there is
        # no need to print anything
        pass


class SitePairSpinSplittingCalculator(SiteMatrixElementCalculator):
    """Class for calculating site pair spin splittings.

    The site pair spin splitting corresponds to a site matrix element with
    f(r) = - 2 W_xc^z(r):

    Δ^(xc,ap)_(nks,n'k+qs') = - 2 <ψ_nks|Θ(r∊Ω_ap)W_xc^z(r)|ψ_n'k+qs'>
    """

    def add_f(self, gd, n_sx, f_x):
        f_x[:] += - 2. * calculate_LSDA_Wxc(gd, n_sx)



from ase import Atoms
from ase.units import Hartree
from gpaw import GPAW
from gpaw.test import equal, gen
import gpaw.mpi as mpi

# Generate non-scalar-relativistic setup for Cu:


def test_generic_Cu(in_tmp_dir):
    setup = gen('Cu', scalarrel=False)

    a = 8.0
    c = a / 2
    Cu = Atoms('Cu', [(c, c, c)], magmoms=[1],
               cell=(a, a, a), pbc=0)

    calc = GPAW(mode='fd',
                h=0.2,
                setups={'Cu': setup})
    Cu.calc = calc
    e = Cu.get_potential_energy()
    niter = calc.get_number_of_iterations()

    e_4s_major = calc.get_eigenvalues(spin=0)[5] / Hartree
    e_3d_minor = calc.get_eigenvalues(spin=1)[4] / Hartree
    print(mpi.rank, e_4s_major, e_3d_minor)

    #
    # The reference values are from:
    #
    #   http://physics.nist.gov/PhysRefData/DFTdata/Tables/29Cu.html
    #
    if mpi.rank == 0:
        print(e_4s_major - e_3d_minor, -0.184013 - -0.197109)
        assert abs(e_4s_major - e_3d_minor - (-0.184013 - -0.197109)) < 0.001

        print(e, niter)
        energy_tolerance = 0.0005
        equal(e, -0.271504, energy_tolerance)



import pytest
from ase import Atom, Atoms

from gpaw import GPAW
from gpaw.lrtddft import LrTDDFT
from gpaw.mpi import world


def get_H2(calculator=None):
    """Define H2 and set calculator if given"""
    R = 0.7  # approx. experimental bond length
    a = 3.0
    c = 4.0
    H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                Atom('H', (a / 2, a / 2, (c + R) / 2))],
               cell=(a, a, c))

    if calculator is not None:
        H2.calc = calculator

    return H2


@pytest.mark.lrtddft
def test_io(in_tmp_dir):
    calc = GPAW(mode='fd', xc='PBE', h=0.25, nbands=5, txt=None)
    calc.calculate(get_H2(calc))
    exlst = LrTDDFT(calc, restrict={'eps': 0.4, 'jend': 3})
    assert len(exlst) == 3
    assert exlst.kss.restrict['eps'] == 0.4

    fname = 'lr.dat.gz'
    exlst.write(fname)
    world.barrier()

    lr2 = LrTDDFT.read(fname)
    assert len(lr2) == 3

    lr3 = LrTDDFT.read(fname, restrict={'jend': 2})
    assert len(lr3) == 2
    assert len(lr3.kss) == 2
    assert len(lr3.Om.fullkss) == 3

    lr4 = LrTDDFT.read(fname, restrict={'energy_range': 20})
    assert len(lr4) == 1


@pytest.mark.lrtddft
def test_invocation():
    calc = GPAW(xc='PBE', mode='fd', h=0.25, nbands=5, txt=None)
    H2 = get_H2(calc)
    exlst = LrTDDFT(restrict={'eps': 0.4, 'jend': 3}, txt=None)
    exlst.calculate(H2)
    assert exlst.xc is not None
    assert hasattr(exlst, 'Om')

    # traditional way
    h2 = get_H2(calc)
    h2.get_potential_energy()
    exlst2 = LrTDDFT(calc, restrict={'eps': 0.4, 'jend': 3}, txt=None)
    assert hasattr(exlst2, 'Om')



from __future__ import annotations

import numbers
from typing import Sequence, overload

import numpy as np
from gpaw.core.matrix import Matrix
from gpaw.gpu import cupy as cp
from gpaw.mpi import MPIComm, serial_comm
from gpaw.new import prod, zips
from gpaw.typing import Array1D, ArrayLike1D, Literal
from gpaw.new.c import dH_aii_times_P_ani_gpu


class AtomArraysLayout:
    def __init__(self,
                 shapes: Sequence[int | tuple[int, ...]],
                 atomdist: AtomDistribution | MPIComm = serial_comm,
                 dtype=float,
                 xp=None):
        """Description of layout of atom arrays.

        Parameters
        ----------
        shapes:
            Shapse of arrays - one for each atom.
        atomdist:
            Distribution of atoms.
        dtype:
            Data-type (float or complex).
        """
        self.shape_a = [shape if isinstance(shape, tuple) else (shape,)
                        for shape in shapes]
        if not isinstance(atomdist, AtomDistribution):
            atomdist = AtomDistribution(np.zeros(len(shapes), int), atomdist)
        self.atomdist = atomdist
        self.dtype = np.dtype(dtype)
        self.xp = xp or np

        self.size = sum(prod(shape) for shape in self.shape_a)

        self.myindices = []
        self.mysize = 0
        I1 = 0
        for a in atomdist.indices:
            I2 = I1 + prod(self.shape_a[a])
            self.myindices.append((a, I1, I2))
            self.mysize += I2 - I1
            I1 = I2

    def __repr__(self):
        return (f'AtomArraysLayout({self.shape_a}, {self.atomdist}, '
                f'{self.dtype}, xp={self.xp.__name__})')

    def new(self, atomdist=None, dtype=None, xp=None):
        """Create new AtomsArrayLayout object with new atomdist."""
        return AtomArraysLayout(self.shape_a,
                                atomdist or self.atomdist,
                                dtype or self.dtype,
                                xp or self.xp)

    def empty(self,
              dims: int | tuple[int, ...] = (),
              comm: MPIComm = serial_comm) -> AtomArrays:
        """Create new AtomArrays object.

        parameters
        ----------
        dims:
            Extra dimensions.
        comm:
            Distribute dimensions along this communicator.
        """
        return AtomArrays(self, dims, comm)

    def zeros(self,
              dims: int | tuple[int, ...] = (),
              comm: MPIComm = serial_comm) -> AtomArrays:
        aa = self.empty(dims, comm)
        aa.data[:] = 0.0
        return aa

    def sizes(self) -> tuple[list[dict[int, int]], Array1D]:
        """Compute array sizes for all ranks.

        >>> AtomArraysLayout([3, 4]).sizes()
        ([{0: 3, 1: 4}], array([7]))
        """
        comm = self.atomdist.comm
        size_ra: list[dict[int, int]] = [{} for _ in range(comm.size)]
        size_r = np.zeros(comm.size, int)
        for a, (rank, shape) in enumerate(zips(self.atomdist.rank_a,
                                               self.shape_a)):
            size = prod(shape)
            size_ra[rank][a] = size
            size_r[rank] += size
        return size_ra, size_r


class AtomDistribution:
    def __init__(self, ranks: ArrayLike1D, comm: MPIComm = serial_comm):
        """Atom-distribution.

        Parameters
        ----------
        ranks:
            List of ranks, one rank per atom.
        comm:
            MPI-communicator.
        """
        self.comm = comm
        self.rank_a = np.array(ranks)
        self.indices = np.where(self.rank_a == comm.rank)[0]

    @classmethod
    def from_number_of_atoms(cls,
                             natoms: int,
                             comm: MPIComm = serial_comm) -> AtomDistribution:
        """Distribute atoms evenly.

        >>> AtomDistribution.from_number_of_atoms(3).rank_a
        array([0, 0, 0])
        """
        blocksize = (natoms + comm.size - 1) // comm.size
        rank_a = np.empty(natoms, int)
        a1 = 0
        for rank in range(comm.size):
            a2 = a1 + blocksize
            rank_a[a1:a2] = rank
            if a2 >= natoms:
                break
            a1 = a2
        return cls(rank_a, comm)

    @classmethod
    def from_atom_indices(cls,
                          atom_indices: Sequence[int],
                          comm: MPIComm = serial_comm,
                          *,
                          natoms: int | None = None) -> AtomDistribution:
        """Create distribution from atom indices.

        >>> AtomDistribution.from_atom_indices([0, 1, 2]).rank_a
        array([0, 0, 0])
        """
        if natoms is None:
            natoms = comm.max(max(atom_indices)) + 1
        rank_a = np.zeros(natoms, int)  # type: ignore
        rank_a[atom_indices] = comm.rank
        comm.sum(rank_a)
        return cls(rank_a, comm)

    def __repr__(self):
        return (f'AtomDistribution(ranks={self.rank_a}, '
                f'comm={self.comm.rank}/{self.comm.size})')

    def gather(self):
        return AtomDistribution(np.zeros(len(self.rank_a), int))


class AtomArrays:
    def __init__(self,
                 layout: AtomArraysLayout,
                 dims: int | tuple[int, ...] = (),
                 comm: MPIComm = serial_comm,
                 data: np.ndarray | None = None):
        """AtomArrays object.

        parameters
        ----------
        layout:
            Layout-description.
        dims:
            Extra dimensions.
        comm:
            Distribute dimensions along this communicator.
        data:
            Data array for storage.
        """
        myshape = (layout.mysize,)
        domain_comm = layout.atomdist.comm
        dtype = layout.dtype

        self.myshape = myshape
        self.comm = comm
        self.domain_comm = domain_comm

        # convert int to tuple:
        self.dims = dims if isinstance(dims, tuple) else (dims,)

        if self.dims:
            mydims0 = (self.dims[0] + comm.size - 1) // comm.size
            d1 = min(comm.rank * mydims0, self.dims[0])
            d2 = min((comm.rank + 1) * mydims0, self.dims[0])
            mydims0 = d2 - d1
            self.mydims = (mydims0,) + self.dims[1:]
        else:
            self.mydims = ()

        fullshape = self.mydims + self.myshape

        if data is not None:
            if data.shape != fullshape:
                raise ValueError(
                    f'Bad shape for data: {data.shape} != {fullshape}')
            if data.dtype != dtype:
                raise ValueError(
                    f'Bad dtype for data: {data.dtype} != {dtype}')
        else:
            data = layout.xp.empty(fullshape, dtype)

        self.data = data
        self._matrix: Matrix | None = None  # matrix view

        self.layout = layout
        self._arrays = {}
        for a, I1, I2 in layout.myindices:
            self._arrays[a] = self.data[..., I1:I2].reshape(
                self.mydims + layout.shape_a[a])
        self.natoms: int = len(layout.shape_a)

    def __repr__(self):
        txt = f'AtomArrays({self.layout}, dims={self.dims}'
        if self.comm.size > 1:
            txt += f', comm={self.comm.rank}/{self.comm.size}'
        return txt + ')'

    @property
    def matrix(self) -> Matrix:
        if self._matrix is not None:
            return self._matrix

        shape = (self.dims[0], prod(self.dims[1:]) * prod(self.myshape))
        myshape = (self.mydims[0], prod(self.mydims[1:]) * prod(self.myshape))
        dist = (self.comm, -1, 1)

        data = self.data.reshape(myshape)
        self._matrix = Matrix(*shape, data=data, dist=dist)

        return self._matrix

    def new(self, *, layout=None, data=None, xp=None):
        """Create new AtomArrays object of same kind.

        Parameters
        ----------
        layout:
            Layout-description.
        data:
            Array to use for storage.
        """
        if xp is np:
            assert layout is None
            assert data is None
            assert self.layout.xp is cp
            layout = self.layout.new(xp=np)
        return AtomArrays(layout or self.layout,
                          self.dims,
                          self.comm,
                          data=data)

    def to_cpu(self):
        if self.layout.xp is np:
            return self
        return self.new(layout=self.layout.new(xp=np),
                        data=cp.asnumpy(self.data))

    def to_xp(self, xp):
        if self.layout.xp is xp:
            assert xp is np, 'cp -> cp should not be needed!'
            return self
        if xp is np:
            return self.new(layout=self.layout.new(xp=np),
                            data=cp.asnumpy(self.data))
        return self.new(layout=self.layout.new(xp=cp),
                        data=cp.asarray(self.data))

    def __getitem__(self, a):
        if isinstance(a, numbers.Integral):
            return self._arrays[a]
        if len(self.dims) == 1:
            a_ai = AtomArrays(self.layout, data=self.data[a[1]].copy())
            return a_ai
        1 / 0

    def get(self, a):
        return self._arrays.get(a)

    def __setitem__(self, a, value):
        self._arrays[a][:] = value

    def __contains__(self, a):
        return a in self._arrays

    def items(self):
        return self._arrays.items()

    def keys(self):
        return self._arrays.keys()

    def values(self):
        return self._arrays.values()

    @overload
    def gather(self, broadcast: Literal[False] = False, copy: bool = False
               ) -> AtomArrays | None:
        ...

    @overload
    def gather(self, broadcast: Literal[True], copy: bool = False
               ) -> AtomArrays:
        ...

    def gather(self, broadcast=False, copy=False):
        """Gather all atoms on master."""
        comm = self.layout.atomdist.comm
        if comm.size == 1:
            if copy:
                aa = self.new()
                aa.data[:] = self.data
                return aa
            return self

        if comm.rank == 0 or broadcast:
            aa = self.new(layout=self.layout.new(atomdist=serial_comm))
        else:
            aa = None

        if comm.rank == 0:
            size_ra, size_r = self.layout.sizes()
            shape = self.mydims + (size_r.max(),)
            buffer = self.layout.xp.empty(shape, self.layout.dtype)
            for rank in range(1, comm.size):
                buf = buffer[..., :size_r[rank]]
                comm.receive(buf, rank)
                b1 = 0
                for a, size in size_ra[rank].items():
                    b2 = b1 + size
                    A = aa[a]
                    A[:] = buf[..., b1:b2].reshape(A.shape)
                    b1 = b2
            for a, array in self._arrays.items():
                aa[a] = array
        else:
            comm.send(self.data, 0)

        if broadcast:
            comm.broadcast(aa.data, 0)

        return aa

    def scatter_from(self,
                     data: np.ndarray | AtomArrays | None = None) -> None:
        if isinstance(data, AtomArrays):
            data = data.data
        comm = self.layout.atomdist.comm
        if comm.size == 1:
            self.data[:] = data
            return

        if comm.rank != 0:
            comm.receive(self.data, 0, 42)
            return

        size_ra, size_r = self.layout.sizes()
        aa = self.new(layout=self.layout.new(atomdist=serial_comm),
                      data=data)
        requests = []
        for rank, (totsize, size_a) in enumerate(zips(size_r, size_ra)):
            if rank != 0:
                buf = np.empty(self.mydims + (totsize,), self.layout.dtype)
                b1 = 0
                for a, size in size_a.items():
                    b2 = b1 + size
                    buf[..., b1:b2] = aa[a].reshape(self.mydims + (size,))
                    b1 = b2
                request = comm.send(buf, rank, 42, False)
                # Remember to store a reference to the
                # send buffer (buf) so that is isn't
                # deallocated
                requests.append((request, buf))
            else:
                for a in size_a:
                    self[a] = aa[a]

        for request, _ in requests:
            comm.wait(request)

    def to_lower_triangle(self):
        """Convert `N*N` matrices to `N*(N+1)/2` vectors.

        >>> a = AtomArraysLayout([(3, 3)]).empty()
        >>> a[0][:] = [[11, 12, 13],
        ...            [12, 22, 23],
        ...            [13, 23, 33]]
        >>> a.to_lower_triangle()[0]
        array([11., 12., 22., 13., 23., 33.])
        """
        shape_a = []
        for i1, i2 in self.layout.shape_a:
            assert i1 == i2
            shape_a.append((i1 * (i1 + 1) // 2,))
        xp = self.layout.xp
        layout = AtomArraysLayout(shape_a,
                                  self.layout.atomdist,
                                  dtype=self.layout.dtype,
                                  xp=xp)
        a_axp = layout.empty(self.dims)
        for a_xii, a_xp in zips(self.values(), a_axp.values()):
            i = a_xii.shape[-1]
            L = xp.tril_indices(i)
            for a_p, a_ii in zips(a_xp.reshape((-1, i * (i + 1) // 2)),
                                  a_xii.reshape((-1, i, i))):
                a_p[:] = a_ii[L]

        return a_axp

    def to_full(self):
        r"""Convert `N(N+1)/2` vectors to `N\times N` matrices.

        >>> a = AtomArraysLayout([6]).empty()
        >>> a[0][:] = [1, 2, 3, 4, 5, 6]
        >>> a.to_full()[0]
        array([[1., 2., 4.],
               [2., 3., 5.],
               [4., 5., 6.]])
        """
        shape_a = []
        for (p,) in self.layout.shape_a:
            i = int((2 * p + 0.25)**0.5)
            shape_a.append((i, i))
        layout = AtomArraysLayout(shape_a,
                                  self.layout.atomdist,
                                  self.layout.dtype)
        a_axii = layout.empty(self.dims)
        for a_xp, a_xii in zips(self.values(), a_axii.values()):
            i = a_xii.shape[-1]
            a_xii[(...,) + np.tril_indices(i)] = a_xp
            u = (...,) + np.triu_indices(i, 1)
            a_xii[u] = np.swapaxes(a_xii, -1, -2)[u].conj()
        return a_axii

    def moved(self, atomdist):
        if (self.layout.atomdist.rank_a == atomdist.rank_a).all():
            return self
        assert self.comm.size == 1
        layout = self.layout.new(atomdist=atomdist)
        new = layout.empty(self.dims)
        comm = atomdist.comm
        requests = []
        for a, I1, I2 in self.layout.myindices:
            r = layout.atomdist.rank_a[a]
            if r == comm.rank:
                new[a][:] = self[a]
            else:
                requests.append(comm.send(self[a], r, block=False))

        for a, I1, I2 in layout.myindices:
            r = self.layout.atomdist.rank_a[a]
            if r != comm.rank:
                comm.receive(new[a], r)

        comm.waitall(requests)
        return new

    def redist(self,
               atomdist: AtomDistribution,
               comm1: MPIComm,
               comm2: MPIComm) -> AtomArrays:
        layout = self.layout.new(atomdist=atomdist)
        result = layout.empty(dims=self.dims)
        if comm1.rank == 0:
            a = self.gather()
        else:
            a = None
        if comm2.rank == 0:
            result.scatter_from(a)
        comm2.broadcast(result.data, 0)
        return result

    def block_diag_multiply(self,
                            block_diag_matrix_axii: AtomArrays,
                            out_ani: AtomArrays,
                            index: int | None = None) -> None:
        """Multiply by block diagonal matrix.

        with A, B and C refering to ``self``, ``block_diag_matrix_axii`` and
        ``out_ani``:::

            --  a   a      a
            >  A   B   -> C
            --  ni  ij     nj
            i

        If index is not None, ``block_diag_matrix_axii`` must have an extra
        dimension: :math:`B_{ij}^{ax}` and x=index is used.
        """
        xp = self.layout.xp
        if xp is np:
            if index is not None:
                block_diag_matrix_axii = block_diag_matrix_axii[:, index]
            for P_ni, dX_ii, out_ni in zips(self.values(),
                                            block_diag_matrix_axii.values(),
                                            out_ani.values()):
                out_ni[:] = P_ni @ dX_ii
            return

        ni_a = xp.array(
            [I2 - I1 for a, I1, I2 in self.layout.myindices],
            dtype=np.int32)
        data = block_diag_matrix_axii.data
        if index is not None:
            data = data[index]
        dH_aii_times_P_ani_gpu(data, ni_a,
                               self.data, out_ani.data)



from os import path

import numpy as np
from ase.utils.timing import Timer
from ase.parallel import parprint

from gpaw import GPAW
from gpaw.fd_operators import Gradient
from gpaw.mpi import world, broadcast, serial_comm
from gpaw.utilities.progressbar import ProgressBar


def get_mml(gs_name='gs.gpw', spin=0, ni=None, nf=None, timer=None):
    """Compute the momentum matrix elements.

    Input:
        gs_name         Ground state file name
        spin            Which spin channel (for spin-polarized systems 0 or 1)
        ni, nf          First and last band to compute the mml (0 to nb)
        timer           Timer to keep track of time
    Output:
        p_kvnn2         A big array in master
    """

    # Start the timer
    if timer is None:
        timer = Timer()
    parprint('Calculating momentum matrix elements...')

    # Load the ground state calculations
    with timer('Load the ground state'):
        parprint('Loading ground state data.')
        assert path.exists(
            gs_name), 'The gs file: {} does not exist!'.format(gs_name)
        calc = GPAW(
            gs_name, txt=None,
            parallel={'kpt': 1, 'band': 1},
            communicator=serial_comm)
        if calc.parameters['mode'] == 'lcao':
            calc.initialize_positions(calc.atoms)

    # Specify desired range and number of bands in calculation
    nb_full = calc.get_number_of_bands()
    ni = int(ni) if ni is not None else 0
    nf = int(nf) if nf is not None else nb_full
    nf = nb_full + nf if (nf < 0) else nf
    blist = list(range(ni, nf))
    nb = len(blist)

    # Spin input
    ns = calc.wfs.nspins
    assert spin < ns, 'Wrong spin input'

    # Real and reciprocal space parameters
    na = len(calc.atoms)
    cell_cv = calc.wfs.gd.cell_cv
    icell_cv = (2 * np.pi) * np.linalg.inv(cell_cv).T
    ibzk_kc = calc.get_ibz_k_points()
    nk = np.shape(ibzk_kc)[0]

    # Parallelisation and memory estimate
    rank = world.rank
    size = world.size
    nkcore = int(np.ceil(nk / size))  # Number of k-points pr. core
    est_mem = 2 * 3 * nk * nb**2 * 16 / 2**20
    parprint('At least {:.2f} MB of memory is required.'.format(est_mem))

    # Allocate the matrix elements
    p_kvnn = np.zeros((nkcore, 3, nb, nb), dtype=complex)

    # if calc.parameters['mode'] == 'lcao':
    nabla_v = [
        Gradient(
            calc.wfs.gd, vv, 1.0, 4,
            complex).apply for vv in range(3)]
    phases = np.ones((3, 2), dtype=complex)

    # Initial call to print 0% progress
    ik = 0
    if rank == 0:
        pb = ProgressBar()

    # Calculate matrix elements in loop over k-points
    for ik in range(nkcore):
        k_ind = rank + size * ik
        if k_ind >= nk:
            break
        k_c = ibzk_kc[k_ind]
        k_v = np.dot(k_c, icell_cv)

        # Get the wavefunctions
        with timer('Get wavefunctions and projections'):
            u_nR = np.array(
                [calc.wfs.get_wave_function_array(
                    ib, k_ind, spin,
                    realspace=True,
                    periodic=True)
                    for ib in blist], complex)

            P_ani = []
            for ia in range(na):
                u = k_ind * ns + spin  # k_ind = q because serial GPAW
                P_ani.append(calc.wfs.kpt_u[u].P_ani[ia][blist])

        # Now compute the momentum part
        grad_nv = calc.wfs.gd.zeros((nb, 3), complex)
        with timer('Momentum calculation'):
            # Get the derivative
            for iv in range(3):
                for ib in range(nb):
                    nabla_v[iv](u_nR[ib], grad_nv[ib, iv], phases)

            # Compute the integral
            p_vnn = np.transpose(
                calc.wfs.gd.integrate(u_nR, grad_nv), (2, 0, 1))

            # Add the overlap part
            M_nn = np.array([calc.wfs.gd.integrate(
                u_nR[ib], u_nR) for ib in range(nb)])
            for iv in range(3):
                p_vnn[iv] += 1j * k_v[iv] * M_nn

        # The PAW corrections are added
        with timer('Add the PAW correction'):
            for ia in range(na):
                setup = calc.wfs.setups[ia]
                P0_ni = P_ani[ia]

                # Loop over components
                for iv in range(3):
                    tmp = setup.nabla_iiv[:, :, iv]
                    p_vnn[iv] += np.dot(
                        np.dot(P0_ni.conj(), tmp), P0_ni.T)

        # Make it momentum and store it
        p_kvnn[ik] = -1j * p_vnn

        # Print the progress
        if rank == 0:
            pb.update(ik / nkcore)
        ik += 1

    if rank == 0:
        pb.finish()

    # Gather all data to the master
    p_kvnn2 = []
    with timer('Gather the data to master'):
        parprint('Gathering date to the master.')
        recv_buf = None
        if rank == 0:
            recv_buf = np.empty((size, nkcore, 3, nb, nb),
                                dtype=complex)
        world.gather(p_kvnn, 0, recv_buf)
        if rank == 0:
            p_kvnn2 = np.zeros((nk, 3, nb, nb), dtype=complex)
            for ii in range(size):
                k_inds = range(ii, nk, size)
                p_kvnn2[k_inds] = recv_buf[ii, :len(k_inds)]

    # Print the timing
    if rank == 0:
        timer.write()

    return p_kvnn2


def make_nlodata(gs_name: str = 'gs.gpw',
                 out_name: str = 'mml.npz',
                 spin: str = 'all',
                 ni: int = 0,
                 nf: int = 0) -> None:
    """Get all required NLO data and store it in a file.

    Writes NLO data to file: w_sk, f_skn, E_skn, p_skvnn.

    Parameters:

    gs_name:
        Ground state file name
    out_name:
        Output filename
    spin:
        Which spin channel ('all', 's0' , 's1')
    ni:
        First band to compute the mml.
    nf:
        Last band to compute the mml (relative to number of bands
        for nf <= 0).

    """

    # Start the timer
    timer = Timer()

    # Get the energy and fermi levels (data is only in master)
    with timer('Get energies and fermi levels'):
        if world.rank == 0:
            # Load the ground state calculation
            calc = GPAW(gs_name, txt=None, communicator=serial_comm)

            # check the GS
            assert not calc.symmetry.point_group, \
                'Point group symmtery should be off.'

            nb_full = calc.get_number_of_bands()
            if nf <= 0:
                nf += nb_full
            ns = calc.wfs.nspins
            if spin == 'all':
                spins = list(range(ns))
            elif spin == 's0':
                spins = [0]
            elif spin == 's1':
                spins = [1]
                assert spins[0] < ns, 'Wrong spin input'
            else:
                raise NotImplementedError

            # Get the data
            w_sk = np.array([calc.get_k_point_weights() for s1 in spins])
            bz_vol = np.linalg.det(2 * np.pi * calc.wfs.gd.icell_cv)
            nk = len(w_sk[0])
            E_skn = np.array([calc.band_structure().todict()['energies'][s1]
                              for s1 in spins])
            f_skn = np.zeros((len(spins), nk, nb_full), dtype=float)
            for sind, s1 in enumerate(spins):
                for ik in range(nk):
                    f_skn[sind, ik] = calc.get_occupation_numbers(
                        kpt=ik, spin=s1) / w_sk[sind, ik] * ns / 2.0
            w_sk *= bz_vol * 2.0 / ns
            broadcast(nf, root=0)
            broadcast(spins, root=0)
        else:
            nf = broadcast(None, root=0)
            spins = broadcast(None, root=0)

    # Compute the momentum matrix elements
    with timer('Compute the momentum matrix elements'):
        p_skvnn = []
        for s1 in spins:
            p_kvnn = get_mml(gs_name=gs_name, spin=s1,
                             ni=ni, nf=nf, timer=timer)
            p_skvnn.append(p_kvnn)

    # Save the output to the file
    if world.rank == 0:
        np.savez(out_name, w_sk=w_sk, f_skn=f_skn[:, :, ni:nf],
                 E_skn=E_skn[:, :, ni:nf], p_skvnn=np.array(p_skvnn, complex))


def get_rml(E_n, p_vnn, pol_v, Etol=1e-6):
    """
    Compute the position matrix elements

    Input:
        E_n             Energies
        p_vnn           Momentum matrix elements
        pol_v           Tensor element
        Etol            Tol. in energy to consider degeneracy
    Output:
        r_vnn, D_vnn    Position and velocity difference matrix el.
    """

    # Useful variables
    nb = len(E_n)
    r_vnn = np.zeros((3, nb, nb), complex)
    D_vnn = np.zeros((3, nb, nb), complex)
    E_nn = np.tile(E_n[:, None], (1, nb)) - \
        np.tile(E_n[None, :], (nb, 1))
    zeroind = np.abs(E_nn) < Etol
    E_nn[zeroind] = 1
    # Loop over components
    for v1 in set(pol_v):
        r_vnn[v1] = p_vnn[v1] / (1j * E_nn)
        r_vnn[v1, zeroind] = 0
        p_n = np.diag(p_vnn[v1])
        D_vnn[v1] = np.tile(p_n[:, None], (1, nb)) - \
            np.tile(p_n[None, :], (nb, 1))

    return r_vnn, D_vnn


def get_derivative(E_n, r_vnn, D_vnn, pol_v, Etol=1e-6):
    """
    Compute the generalized derivative of position matrix elements

    Input:
        E_n             Energies
        r_vnn           Momentum matrix elements
        D_vnn           Velocity difference
        pol_v           Tensor element
        Etol            Tol. in energy to consider degeneracy
    Output:
        rd_vvnn         Generilized derivative of position
    """

    # Useful variables
    nb = len(E_n)
    rd_vvnn = np.zeros((3, 3, nb, nb), complex)
    E_nn = np.tile(E_n[:, None], (1, nb)) - \
        np.tile(E_n[None, :], (nb, 1))
    zeroind = np.abs(E_nn) < Etol
    E_nn[zeroind] = 1
    for v1 in set(pol_v):
        for v2 in set(pol_v):
            tmp = (r_vnn[v1] * np.transpose(D_vnn[v2])
                   + r_vnn[v2] * np.transpose(D_vnn[v1])
                   + 1j * np.dot(r_vnn[v1], r_vnn[v2] * E_nn)
                   - 1j * np.dot(r_vnn[v2] * E_nn, r_vnn[v1])) / E_nn
            tmp[zeroind] = 0
            rd_vvnn[v1, v2] = tmp

    return rd_vvnn



from __future__ import annotations

from gpaw.core import UGDesc
from gpaw.core.arrays import DistributedArrays as XArray
from gpaw.core.atom_arrays import AtomDistribution
from gpaw.core.uniform_grid import UGArray
from gpaw.fd_operators import Gradient, Laplace
from gpaw.new import cached_property, zips
from gpaw.new.builder import create_uniform_grid
from gpaw.new.fd.pot_calc import FDPotentialCalculator
from gpaw.new.hamiltonian import Hamiltonian
from gpaw.new.poisson import PoissonSolver, PoissonSolverWrapper
from gpaw.new.pwfd.builder import PWFDDFTComponentsBuilder
from gpaw.poisson import PoissonSolver as make_poisson_solver


class FDDFTComponentsBuilder(PWFDDFTComponentsBuilder):
    def __init__(self, atoms, params, *, comm, nn=3, interpolation=3):
        super().__init__(atoms,
                         params,
                         comm=comm)
        assert not self.soc
        self.kin_stencil_range = nn
        self.interpolation_stencil_range = interpolation

        self._nct_aR = None
        self._tauct_aR = None

        self.electrostatic_potential_desc = self.fine_grid
        self.interpolation_desc = self.fine_grid

    @cached_property
    def atomdist(self) -> AtomDistribution:
        return AtomDistribution(
            self.grid.ranks_from_fractional_positions(self.fracpos_ac),
            self.grid.comm)

    def create_uniform_grids(self):
        grid = create_uniform_grid(
            self.mode,
            self.params.gpts,
            self.atoms.cell,
            self.atoms.pbc,
            self.ibz.symmetries,
            h=self.params.h,
            interpolation='not fft',
            comm=self.communicators['d'])
        fine_grid = grid.new(size=grid.size_c * 2)
        # decomposition=[2 * d for d in grid.decomposition]
        return grid, fine_grid

    def create_wf_description(self) -> UGDesc:
        return self.grid.new(dtype=self.dtype)

    def get_pseudo_core_densities(self):
        if self._nct_aR is None:
            self._nct_aR = self.setups.create_pseudo_core_densities(
                self.grid, self.fracpos_ac, atomdist=self.atomdist, xp=self.xp)
        return self._nct_aR

    def get_pseudo_core_ked(self):
        if self._tauct_aR is None:
            self._tauct_aR = self.setups.create_pseudo_core_ked(
                self.grid, self.fracpos_ac, atomdist=self.atomdist)
        return self._tauct_aR

    def create_poisson_solver(self) -> PoissonSolver:
        solver = make_poisson_solver(**self.params.poissonsolver)
        solver.set_grid_descriptor(self.fine_grid._gd)
        return PoissonSolverWrapper(solver)

    def create_potential_calculator(self):
        poisson_solver = self.create_poisson_solver()
        return FDPotentialCalculator(
            self.grid, self.fine_grid, self.setups, self.xc, poisson_solver,
            fracpos_ac=self.fracpos_ac, atomdist=self.atomdist,
            interpolation_stencil_range=self.interpolation_stencil_range,
            xp=self.xp)

    def create_hamiltonian_operator(self, blocksize=10):
        return FDHamiltonian(self.wf_desc, self.kin_stencil_range, blocksize)

    def convert_wave_functions_from_uniform_grid(self,
                                                 C_nM,
                                                 basis_set,
                                                 kpt_c,
                                                 q):
        grid = self.grid.new(kpt=kpt_c, dtype=self.dtype)
        psit_nR = grid.zeros(self.nbands, self.communicators['b'])
        mynbands = len(C_nM.data)
        basis_set.lcao_to_grid(C_nM.data, psit_nR.data[:mynbands], q)
        return psit_nR

    def read_ibz_wave_functions(self, reader):
        ibzwfs = super().read_ibz_wave_functions(reader)

        if 'coefficients' in reader.wave_functions:
            name = 'coefficients'
        elif 'values' in reader.wave_functions:
            name = 'values'
        else:
            return ibzwfs

        c = reader.bohr**1.5
        if reader.version < 0:
            c = 1  # old gpw file

        for wfs in ibzwfs:
            grid = self.wf_desc.new(kpt=wfs.kpt_c)
            index = (wfs.spin, wfs.k)
            data = reader.wave_functions.proxy(name, *index)
            data.scale = c
            if self.communicators['w'].size == 1:
                wfs.psit_nX = UGArray(grid, self.nbands, data=data)
            else:
                band_comm = self.communicators['b']
                wfs.psit_nX = UGArray(
                    grid, self.nbands,
                    comm=band_comm)
                if grid.comm.rank == 0:
                    mynbands = (self.nbands +
                                band_comm.size - 1) // band_comm.size
                    n1 = min(band_comm.rank * mynbands, self.nbands)
                    n2 = min((band_comm.rank + 1) * mynbands, self.nbands)
                    assert wfs.psit_nX.mydims[0] == n2 - n1
                    data = data[n1:n2]  # read from file
                wfs.psit_nX.scatter_from(data)

        return ibzwfs


class FDHamiltonian(Hamiltonian):
    def __init__(self, grid, kin_stencil=3, blocksize=10):
        self.grid = grid
        self.blocksize = blocksize
        self._gd = grid._gd
        self.kin = Laplace(self._gd, -0.5, kin_stencil, grid.dtype)

        # For MGGA:
        self.grad_v = []

    def apply_local_potential(self,
                              vt_R: UGArray,
                              psit_nR: XArray,
                              out: XArray,
                              ) -> None:
        assert isinstance(psit_nR, UGArray)
        assert isinstance(out, UGArray)
        self.kin(psit_nR, out)
        for p, o in zips(psit_nR.data, out.data):
            o += p * vt_R.data

    def apply_mgga(self,
                   dedtaut_R: UGArray,
                   psit_nR: XArray,
                   vt_nR: XArray) -> None:
        if len(self.grad_v) == 0:
            grid = psit_nR.desc
            self.grad_v = [
                Gradient(grid._gd, v, n=3, dtype=grid.dtype)
                for v in range(3)]

        tmp_R = psit_nR.desc.empty()
        for psit_R, out_R in zips(psit_nR, vt_nR):
            for grad in self.grad_v:
                grad(psit_R, tmp_R)
                tmp_R.data *= dedtaut_R.data
                grad(tmp_R, tmp_R)
                tmp_R.data *= 0.5
                out_R.data -= tmp_R.data

    def create_preconditioner(self, blocksize):
        from types import SimpleNamespace

        from gpaw.preconditioner import Preconditioner as PC
        pc = PC(self._gd, self.kin, self.grid.dtype, self.blocksize)

        def apply(psit, residuals, out):
            kpt = SimpleNamespace(phase_cd=psit.desc.phase_factor_cd)
            pc(residuals.data, kpt, out=out.data)

        return apply



import sys
from optparse import OptionParser

from gpaw.atom.basis import BasisMaker
from gpaw.atom.configurations import parameters, parameters_extra
from gpaw.setup_data import SetupData
from gpaw.mpi import world

# Module for generating basis sets more suitable for excited states.
#
# Generates basis sets that include empty p orbitals as valence states
# for metals.  This improves the description of excited states and is
# useful for TDDFT, band structures (the part above the Fermi level),
# and more.


class BasisSpecification:
    def __init__(self, setup, jvalues, jextra):
        self.setup = setup
        self.jvalues = jvalues
        self.jextra = jextra

    def __str__(self):
        l_j = self.setup.l_j
        jvtxt = ', '.join(['%s(l=%s)' % (j, l_j[j]) for j in self.jvalues])
        jetxt = ', '.join(['%s(l=%s)' % (j, l_j[j]) for j in self.jextra])
        return '%s: jvalues=[%s], jextra=[%s]' % (self.setup.symbol,
                                                  jvtxt, jetxt)


description = """Generate basis sets that include unoccupied p states as
valence states instead of Gaussian-based polarization functions.
If SYMBOLs are omitted, generate basis sets for all elements with
unoccupied p states and default setup."""


def main():
    parser = OptionParser(usage='%prog [OPTION...] [SYMBOL...]',
                          description=description)
    parser.add_option('--xc', metavar='FUNCTIONAL',
                      default='PBE',
                      help='generate basis sets for FUNCTIONAL[=%default]')
    parser.add_option('--from', metavar='SYMBOL', dest='_from',
                      help='generate starting from SYMBOL if generating '
                      'for all elements')
    opts, symbols = parser.parse_args()

    if len(symbols) == 0:
        symbols = sorted(parameters.keys())
        othersymbols = []
        for symbol in parameters_extra:
            name = parameters_extra[symbol]['name']
            code = '%s.%s' % (symbol, name)
            othersymbols.append(code)
        # Setups that cause trouble
        # trouble = set(['Os.8', 'Ta.5', 'V.5', 'W.6', 'Ir.9'])
        trouble = set([])
        othersymbols = [symbol for symbol in othersymbols
                        if symbol not in trouble]
        symbols.extend(sorted(othersymbols))

        if opts._from:
            index = symbols.index(opts._from)
            symbols = symbols[index:]

    specifications = []
    for sym in symbols:
        try:
            s = SetupData(sym, opts.xc)
        except RuntimeError as e:
            if str(e).startswith('Could not find'):
                # print 'No %s' % sym
                continue
            else:
                raise

        # One could include basis functions also for the ``virtual'' states
        # (marked with negative n)

        jvalues = []
        jextra = []
        for j in range(len(s.f_j)):
            if s.eps_j[j] < 0:
                jvalues.append(j)
                if s.f_j[j] == 0.0 and s.n_j[j] > 0:
                    jextra.append(j)
        if len(jextra) > 0:
            specifications.append(BasisSpecification(s, jvalues, jextra))
            # print sym, jvalues
        # XXX check whether automatic settings coincide with those of official
        # setups distribution
        # bm = BasisMaker(sym, ''

    if world.rank == 0:
        print('Generating basis sets for: %s'
              % ' '.join(spec.setup.symbol for spec in specifications))
    sys.stdout.flush()
    world.barrier()

    for i, spec in enumerate(specifications):
        if i % world.size != world.rank:
            continue
        if world.size > 1:
            print(world.rank, spec)
        else:
            print(spec)
        gtxt = None

        tokens = spec.setup.symbol.split('.')
        sym = tokens[0]

        if len(tokens) == 1:
            p = parameters
            name = 'pvalence'
        elif len(tokens) == 2:
            p = parameters_extra
            name = '%s.pvalence' % tokens[1]
        else:
            raise ValueError('Strange setup specification')

        # This generates only dz setups
        bm = BasisMaker(sym, '%s' % (name),
                        run=False, gtxt=gtxt, xc=opts.xc)
        bm.generator.run(write_xml=False, **p[sym])
        basis = bm.generate(2, 0, txt=None, jvalues=spec.jvalues)
        basis.write_xml()


if __name__ == '__main__':
    main()



import numpy as np
import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.fdtd.poisson_fdtd import FDTDPoissonSolver
from gpaw.fdtd.polarizable_material import (PermittivityPlus,
                                            PolarizableMaterial,
                                            PolarizableSphere)
from gpaw.inducedfield.inducedfield_base import BaseInducedField
from gpaw.inducedfield.inducedfield_fdtd import (
    FDTDInducedField, calculate_hybrid_induced_field)
from gpaw.inducedfield.inducedfield_tddft import TDDFTInducedField
from gpaw.mpi import world
from gpaw.tddft import TDDFT, DipoleMomentWriter
from gpaw.test import equal


@pytest.mark.later
def test_fdtd_ed_inducedfield(in_tmp_dir):
    do_print_values = 0  # Use this for printing the reference values

    # Accuracy
    energy_eps = 0.0005
    poisson_eps = 1e-12
    density_eps = 1e-6

    # Whole simulation cell (Angstroms)
    large_cell = [20, 20, 30]

    # Quantum subsystem
    atom_center = np.array([10.0, 10.0, 20.0])
    atoms = Atoms('Na2', [atom_center + [0.0, 0.0, -1.50],
                          atom_center + [0.0, 0.0, +1.50]])

    # Permittivity file
    if world.rank == 0:
        fo = open('ed.txt', 'w')
        fo.writelines(['1.20 0.20 25.0'])
        fo.close()
    world.barrier()

    # Classical subsystem
    classical_material = PolarizableMaterial()
    sphere_center = np.array([10.0, 10.0, 10.0])
    classical_material.add_component(
        PolarizableSphere(permittivity=PermittivityPlus('ed.txt'),
                          center=sphere_center,
                          radius=5.0))

    # Combined Poisson solver
    poissonsolver = FDTDPoissonSolver(classical_material=classical_material,
                                      eps=poisson_eps,
                                      qm_spacing=0.40,
                                      cl_spacing=0.40 * 4,
                                      cell=large_cell,
                                      remove_moments=(1, 4),
                                      communicator=world,
                                      potential_coupler='Refiner')
    poissonsolver.set_calculation_mode('iterate')

    # Combined system
    atoms.set_cell(large_cell)
    atoms, qm_spacing, gpts = poissonsolver.cut_cell(atoms,
                                                     vacuum=2.50)

    # Initialize GPAW
    gs_calc = GPAW(mode='fd',
                   gpts=gpts,
                   eigensolver='cg',
                   nbands=-1,
                   poissonsolver=poissonsolver,
                   symmetry={'point_group': False},
                   convergence={'energy': energy_eps,
                                'density': density_eps})
    atoms.calc = gs_calc

    # Ground state
    energy = atoms.get_potential_energy()

    # Test ground state
    equal(energy, -0.631881, energy_eps * gs_calc.get_number_of_electrons())

    # Test floating point arithmetic errors
    equal(gs_calc.hamiltonian.poisson.shift_indices_1, [4, 4, 10], 0)
    equal(gs_calc.hamiltonian.poisson.shift_indices_2, [8, 8, 16], 0)

    # Save state
    gs_calc.write('gs.gpw', 'all')
    classical_material = None
    gs_calc = None

    # Initialize TDDFT and FDTD
    kick = [0.0, 0.0, 1.0e-3]
    time_step = 10.0
    iterations = 10

    td_calc = TDDFT('gs.gpw')
    DipoleMomentWriter(td_calc, 'dm.dat')

    # Attach InducedFields to the calculation
    frequencies = [2.05, 4.0]
    width = 0.15
    cl_ind = FDTDInducedField(paw=td_calc,
                              frequencies=frequencies,
                              width=width)
    qm_ind = TDDFTInducedField(paw=td_calc,
                               frequencies=frequencies,
                               width=width)

    # Propagate TDDFT and FDTD
    td_calc.absorption_kick(kick_strength=kick)
    td_calc.propagate(time_step, iterations // 2)
    td_calc.write('td.gpw', 'all')
    cl_ind.write('cl.ind')
    qm_ind.write('qm.ind')
    td_calc = None
    cl_ind.paw = None
    qm_ind.paw = None
    cl_ind = None
    qm_ind = None

    # Restart
    td_calc = TDDFT('td.gpw')
    DipoleMomentWriter(td_calc, 'dm.dat')
    cl_ind = FDTDInducedField(filename='cl.ind',
                              paw=td_calc)
    qm_ind = TDDFTInducedField(filename='qm.ind',
                               paw=td_calc)
    td_calc.propagate(time_step, iterations // 2)
    td_calc.write('td.gpw', 'all')
    cl_ind.write('cl.ind')
    qm_ind.write('qm.ind')

    # Test
    ref_cl_dipole_moment = [5.25374117e-14, 5.75811267e-14, 3.08349334e-02]
    ref_qm_dipole_moment = [1.78620337e-11, -1.57782578e-11, 5.21368300e-01]

    tol = 1e-4
    equal(td_calc.hamiltonian.poisson.get_classical_dipole_moment(),
          ref_cl_dipole_moment, tol)
    equal(td_calc.hamiltonian.poisson.get_quantum_dipole_moment(),
          ref_qm_dipole_moment, tol)

    cl_ind.paw = None
    qm_ind.paw = None
    td_calc = None
    cl_ind = None
    qm_ind = None

    # Calculate induced fields
    td_calc = TDDFT('td.gpw')

    # Classical subsystem
    cl_ind = FDTDInducedField(filename='cl.ind', paw=td_calc)
    cl_ind.calculate_induced_field(gridrefinement=2)
    cl_ind.write('cl_field.ind', mode='all')

    # Quantum subsystem
    qm_ind = TDDFTInducedField(filename='qm.ind', paw=td_calc)
    qm_ind.calculate_induced_field(gridrefinement=2)
    qm_ind.write('qm_field.ind', mode='all')

    # Total system, interpolate/extrapolate to a grid with spacing h
    tot_ind = calculate_hybrid_induced_field(cl_ind, qm_ind, h=0.4)
    tot_ind.write('tot_field.ind', mode='all')

    tot_ind.paw = None
    cl_ind.paw = None
    qm_ind.paw = None
    td_calc = None
    cl_ind = None
    qm_ind = None
    tot_ind = None

    # Test induced fields
    ref_values = [72404.467117024149,
                  0.520770766296,
                  0.520770766299,
                  0.830247064075,
                  72416.234345610734,
                  0.517294132489,
                  0.517294132492,
                  0.824704513888,
                  2451.767847927681,
                  0.088037476748,
                  0.088037476316,
                  0.123334033914,
                  2454.462292798476,
                  0.087537484422,
                  0.087537483971,
                  0.122592730690,
                  76582.089818637178,
                  0.589941751987,
                  0.589941751804,
                  0.869526245360,
                  76592.175846021099,
                  0.586223386358,
                  0.586223386102,
                  0.864478308364]

    for fname in ['cl_field.ind', 'qm_field.ind', 'tot_field.ind']:
        ind = BaseInducedField(filename=fname,
                               readmode='field')
        # Estimate tolerance (worst case error accumulation)
        tol = (iterations * ind.fieldgd.integrate(ind.fieldgd.zeros() + 1.0) *
               max(density_eps, np.sqrt(poisson_eps)))
        if do_print_values:
            print('tol = %.12f' % tol)
        for w in range(len(frequencies)):
            val = ind.fieldgd.integrate(ind.Ffe_wg[w])
            equal(val, ref_values.pop(0), tol)
            for v in range(3):
                val = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[w][v]))
                equal(val, ref_values.pop(0), tol)



from __future__ import annotations

import numpy as np
from gpaw.core.domain import Domain
from gpaw.core.plane_waves import PWDesc
from gpaw.typing import Vector


class SpinorWaveFunctionDescriptor(Domain):
    def __init__(self,
                 pw: PWDesc,
                 qspiral_v: Vector | None = None):
        self.pw = pw
        self.qspiral_v = (np.asarray(qspiral_v) if qspiral_v is not None else
                          None)
        Domain.__init__(self, pw.cell_cv, pw.pbc_c, pw.kpt_c, pw.comm,
                        complex)
        self.myshape = (2,) + pw.myshape
        self.itemsize = pw.itemsize
        self.shape = (2,) + pw.shape
        self.dv = pw.dv

    def __repr__(self):
        q = self.qspiral_v
        return f'{self.__class__.__name__}({self.pw}, qspiral_v={q})'

    def new(self, *, kpt):
        pw = self.pw.new(kpt=kpt)
        pw.qspiral_v = self.qspiral_v
        return SpinorWaveFunctionDescriptor(pw, self.qspiral_v)

    def empty(self, shape, comm, xp=None):
        assert isinstance(shape, int)
        return self.pw.empty((shape, 2), comm)

    def global_shape(self) -> tuple[int, ...]:
        return (2,) + self.pw.global_shape()

    def indices(self, size):
        return self.pw.indices(size)



import numpy as np
from gpaw import GPAW
from gpaw.berryphase import get_polarization_phase, parallel_transport
from gpaw.berryphase import get_berry_phases
import gpaw.mpi as mpi
import pytest


# Values from an earlier test
ref_phi_mos2_km = np.array(
    [[2.72907676e-04, 2.99369724e+00, 4.51932187e+00, 5.94725651e+00],
     [4.84334561e-03, 2.42519044e+00, 4.43335136e+00, 5.75115262e+00],
     [2.99682618e-02, 2.26119678e+00, 4.30480687e+00, 5.78042986e+00],
     [4.84334561e-03, 2.42519044e+00, 4.43335136e+00, 5.75115262e+00],
     [2.72907676e-04, 2.99369724e+00, 4.51932187e+00, 5.94725651e+00],
     [3.75847658e-03, 2.67197983e+00, 4.36511629e+00, 5.60446187e+00]])


@pytest.mark.later
def test_parallel_transport_mos2(in_tmp_dir, gpw_files):
    # Calculate the berry phases and spin projections
    gpw = gpw_files['mos2_pw_nosym']
    parallel_transport(str(gpw), name='mos2', scale=1)

    # Load phase-ordered data
    phi_km, S_km = load_renormalized_data('mos2')

    # Test that the berry phases do not change (assuming that they
    # were correct to begin with)
    print(phi_km[:, ::7])  # we slice the bands to make output readable
    assert phi_km[:, ::7] == pytest.approx(ref_phi_mos2_km, abs=0.05)


@pytest.mark.later
def test_parallel_transport_i2sb2(in_tmp_dir, gpw_files):
    # Calculate the berry phases and spin projections
    calc = GPAW(gpw_files['i2sb2_pw_nosym'],
                txt=None, communicator=mpi.serial_comm)
    nelec = int(calc.get_number_of_electrons())
    parallel_transport(calc, name='i2sb2', scale=1,
                       # To calculate the valence bands berry
                       # phases, we only need the top valence
                       # group of bands. This corresponds to 2x8
                       # bands, see c2db (x2 for spin)
                       bands=range(nelec - 2 * 8, nelec))

    # Load phase-ordered data
    phi_km, S_km = load_renormalized_data('i2sb2')

    # # For the spin test below to make sense, please compare this
    # # plot to the berry phase plot at the c2db website
    # import matplotlib.pyplot as plt
    # plt.scatter(np.tile(np.arange(len(phi_km)), len(phi_km.T)),
    #             phi_km.T.reshape(-1),
    #             cmap=plt.get_cmap('viridis'),
    #             c=S_km.T.reshape(-1),
    #             s=25,
    #             marker='o')
    # plt.ylim((0, 2 * np.pi))
    # plt.show()

    # We test the spin for bands we are in control of, that is,
    # avoid high-symmetry points and look only at the winding
    # bands above a phase of ~pi, see the c2db berry phase plot
    bands = [0, 1, 3, 4]
    phi_qm = phi_km[bands]
    S_qm = S_km[bands]
    Svalues = S_qm[phi_qm > 3.0]
    assert Svalues == pytest.approx(np.array([-1, 1,  # k=0
                                              -1, 1,  # k=1
                                              1, -1,  # k=2
                                              1, -1]),   # k=3
                                    abs=0.01)
    # Test also the berry phases for the same bands
    phivalues = phi_qm[phi_qm > 3.0]
    print(phivalues)
    # We test that the values don't change too much. This will
    # also guarantee that the results agree qualitatively with
    # the c2db plot
    assert phivalues == pytest.approx(
        [3.115, 5.309, 3.970, 4.455,
         3.970, 4.455, 3.115, 5.309], abs=0.05)


def load_renormalized_data(name):
    data = np.load(f'phases_{name}.npz')
    phi_km = data['phi_km']
    S_km = data['S_km']

    # Phases are only well-defined modulo 2pi
    phi_km %= 2 * np.pi

    # Sort bands by the berry phase
    indices = np.argsort(phi_km, axis=1)
    phi_km = np.take_along_axis(phi_km, indices, axis=1)
    S_km = np.take_along_axis(S_km, indices, axis=1)

    return phi_km, S_km


def test_pol(in_tmp_dir, gpw_files):

    # It is ugly to convert to string. But this is required in
    # get_polarization_phase. Should be changed in the future...
    phi_c = get_polarization_phase(str(gpw_files['mos2_pw_nosym']))

    # Only should test modulo 2pi
    phi = phi_c / (2 * np.pi)
    phitest = [6.60376287e-01, 3.39625036e-01, 0.0]
    err = phi - phitest
    assert err == pytest.approx(err.round(), abs=1e-3)


def test_berry_phases(in_tmp_dir, gpw_files):

    calc = GPAW(gpw_files['mos2_pw_nosym'],
                communicator=mpi.serial_comm)

    ind, phases = get_berry_phases(calc)

    indtest = [[0, 6, 12, 18, 24, 30],
               [1, 7, 13, 19, 25, 31],
               [2, 8, 14, 20, 26, 32],
               [3, 9, 15, 21, 27, 33],
               [4, 10, 16, 22, 28, 34],
               [5, 11, 17, 23, 29, 35]]

    phasetest = [1.66179, 2.54985, 3.10069, 2.54985, 1.66179, 0.92385]
    assert np.allclose(ind, indtest)
    assert np.allclose(phases, phasetest, atol=1e-3)


# only master will raise, so this test will hang in parallel
@pytest.mark.serial
def test_assertions(in_tmp_dir, gpw_files):
    """
    Functions should only work without symmetry
    Tests so that proper assertion is raised for calculator
    with symmetry enabled
    """

    gpw_file = gpw_files['mos2_pw']
    with pytest.raises(AssertionError):
        get_polarization_phase(str(gpw_file))

    calc = GPAW(gpw_file,
                communicator=mpi.serial_comm)

    with pytest.raises(AssertionError):
        ind, phases = get_berry_phases(calc)

    with pytest.raises(AssertionError):
        phi_km, S_km = parallel_transport(calc,
                                          direction=0,
                                          name='mos2', scale=0)



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""
Python wrapper functions for the ``C`` package:
Basic Linear Algebra Subroutines (BLAS)

See also:
http://en.wikipedia.org/wiki/Basic_Linear_Algebra_Subprograms
and
http://www.netlib.org/lapack/lug/node145.html
"""
from typing import TypeVar

import _gpaw
import numpy as np
import scipy.linalg.blas as blas
from gpaw import debug
from gpaw.new import prod
from gpaw.typing import Array2D, ArrayND
from gpaw.utilities import is_contiguous


def is_finite(array, tril=False):
    if isinstance(array, np.ndarray):
        xp = np
    else:
        from gpaw.gpu import cupy as xp
    if tril:
        array = xp.tril(array)
    return xp.isfinite(array).all()


__all__ = ['mmm']

T = TypeVar('T', float, complex)


def mmm(alpha: T,
        a: Array2D,
        opa: str,
        b: Array2D,
        opb: str,
        beta: T,
        c: Array2D) -> None:
    """Matrix-matrix multiplication using dgemm or zgemm.

    For opa='N' and opb='N', we have:::

        c <- αab + βc.

    Use 'T' to transpose matrices and 'C' to transpose and complex conjugate
    matrices.
    """

    assert opa in 'NTC'
    assert opb in 'NTC'

    if opa == 'N':
        a1, a2 = a.shape
    else:
        a2, a1 = a.shape
    if opb == 'N':
        b1, b2 = b.shape
    else:
        b2, b1 = b.shape
    assert a2 == b1
    assert c.shape == (a1, b2)

    assert a.dtype == b.dtype == c.dtype
    assert a.strides[1] == c.itemsize or a.size == 0
    assert b.strides[1] == c.itemsize or b.size == 0
    assert c.strides[1] == c.itemsize or c.size == 0
    if a.dtype == float:
        assert not isinstance(alpha, complex)
        assert not isinstance(beta, complex)
    else:
        assert a.dtype == complex

    _gpaw.mmm(alpha, a, opa, b, opb, beta, c)


def gpu_mmm(alpha, a, opa, b, opb, beta, c):
    """Launch CPU or GPU version of mmm()."""
    m = b.shape[1] if opb == 'N' else b.shape[0]
    n = a.shape[0] if opa == 'N' else a.shape[1]
    k = b.shape[0] if opb == 'N' else b.shape[1]
    lda = a.strides[0] // a.itemsize
    ldb = b.strides[0] // b.itemsize
    ldc = c.strides[0] // c.itemsize
    _gpaw.mmm_gpu(alpha, a.data.ptr, lda, opa,
                  b.data.ptr, ldb, opb, beta,
                  c.data.ptr, ldc, c.itemsize,
                  m, n, k)


def gpu_scal(alpha, x):
    """alpha x

    Performs the operation::

      x <- alpha * x

    """
    if debug:
        if isinstance(alpha, complex):
            assert is_contiguous(x, complex)
        else:
            assert isinstance(alpha, float)
            assert x.dtype in [float, complex]
            assert x.flags.c_contiguous
    _gpaw.scal_gpu(alpha, x.data.ptr, x.shape, x.dtype)


def to2d(array: ArrayND) -> Array2D:
    """2D view af ndarray.

    >>> to2d(np.zeros((2, 3, 4))).shape
    (2, 12)
    """
    shape = array.shape
    return array.reshape((shape[0], prod(shape[1:])))


def mmmx(alpha: T,
         a: ArrayND,
         opa: str,
         b: ArrayND,
         opb: str,
         beta: T,
         c: ArrayND) -> None:
    """Matrix-matrix multiplication using dgemm or zgemm.

    Arrays a, b and c are converted to 2D arrays before calling mmm().
    """
    mmm(alpha, to2d(a), opa, to2d(b), opb, beta, to2d(c))


def gpu_gemm(alpha, a, b, beta, c, transa='n'):
    """General Matrix Multiply.

    Performs the operation::

      c <- alpha * b.a + beta * c

    If transa is "n", ``b.a`` denotes the matrix multiplication defined by::

                      _
                     \
      (b.a)        =  ) b  * a
           ijkl...   /_  ip   pjkl...
                      p

    If transa is "t" or "c", ``b.a`` denotes the matrix multiplication
    defined by::

                      _
                     \
      (b.a)        =  ) b    *    a
           ij        /_  iklm...   jklm...
                     klm...

    where in case of "c" also complex conjugate of a is taken.
    """
    if debug:
        assert beta == 0.0 or is_finite(c)

        assert (a.dtype == float and b.dtype == float and c.dtype == float and
                isinstance(alpha, float) and isinstance(beta, float) or
                a.dtype == complex and b.dtype == complex and
                c.dtype == complex)
        assert a.flags.c_contiguous
        if transa == 'n':
            assert c.flags.c_contiguous or (c.ndim == 2
                                            and c.strides[1] == c.itemsize)
            assert b.ndim == 2
            assert b.strides[1] == b.itemsize
            assert a.shape[0] == b.shape[1]
            assert c.shape == b.shape[0:1] + a.shape[1:]
        else:
            assert b.size == 0 or b[0].flags.c_contiguous
            assert c.strides[1] == c.itemsize
            assert a.shape[1:] == b.shape[1:]
            assert c.shape == (b.shape[0], a.shape[0])

    _gpaw.gemm_gpu(alpha, a.data.ptr, a.shape,
                   b.data.ptr, b.shape, beta,
                   c.data.ptr, c.shape,
                   a.dtype, transa)


def gpu_gemv(alpha, a, x, beta, y, trans='t'):
    """General Matrix Vector product.

    Performs the operation::

      y <- alpha * a.x + beta * y

    ``a.x`` denotes matrix multiplication, where the product-sum is
    over the entire length of the vector x and
    the first dimension of a (for trans='n'), or
    the last dimension of a (for trans='t' or 'c').

    If trans='c', the complex conjugate of a is used. The default is
    trans='t', i.e. behaviour like np.dot with a 2D matrix and a vector.
    """
    if debug:
        assert (a.dtype == float and x.dtype == float and y.dtype == float and
                isinstance(alpha, float) and isinstance(beta, float) or
                a.dtype == complex and x.dtype == complex and
                y.dtype == complex)
        assert a.flags.c_contiguous
        assert y.flags.c_contiguous
        assert x.ndim == 1
        assert y.ndim == a.ndim - 1
        if trans == 'n':
            assert a.shape[0] == x.shape[0]
            assert a.shape[1:] == y.shape
        else:
            assert a.shape[-1] == x.shape[0]
            assert a.shape[:-1] == y.shape

    _gpaw.gemv_gpu(alpha, a.data.ptr, a.shape,
                   x.data.ptr, x.shape, beta,
                   y.data.ptr, a.dtype,
                   trans)


def axpy(alpha, x, y):
    """alpha x plus y.

    Performs the operation::

      y <- alpha * x + y

    """
    if x.size == 0:
        return
    assert x.flags.contiguous
    assert y.flags.contiguous
    x = x.ravel()
    y = y.ravel()
    if x.dtype == float:
        z = blas.daxpy(x, y, a=alpha)
    else:
        z = blas.zaxpy(x, y, a=alpha)
    assert z is y, (x, y, x.shape, y.shape)


def gpu_axpy(alpha, x, y):
    """alpha x plus y.

    Performs the operation::

      y <- alpha * x + y

    """
    if debug:
        if isinstance(alpha, complex):
            assert is_contiguous(x, complex) and is_contiguous(y, complex)
        else:
            assert isinstance(alpha, float)
            assert x.dtype in [float, complex]
            assert x.dtype == y.dtype
            assert x.flags.c_contiguous and y.flags.c_contiguous
        assert x.shape == y.shape

    _gpaw.axpy_gpu(alpha, x.data.ptr, x.shape,
                   y.data.ptr, y.shape,
                   x.dtype)


def rk(alpha, a, beta, c, trans='c'):
    """Rank-k update of a matrix.

    For ``trans='c'`` the following operation is performed:::

              †
      c <- αaa + βc,

    and for ``trans='t'`` we get:::

             †
      c <- αa a + βc

    If the ``a`` array has more than 2 dimensions then the 2., 3., ...
    axes are combined.

    Only the lower triangle of ``c`` will contain sensible numbers.
    """
    if debug:
        assert beta == 0.0 or is_finite(c, tril=True)

        assert (a.dtype == float and c.dtype == float or
                a.dtype == complex and c.dtype == complex)
        assert a.flags.c_contiguous
        assert a.ndim > 1
        if trans == 'n':
            assert c.shape == (a.shape[1], a.shape[1])
        else:
            assert c.shape == (a.shape[0], a.shape[0])
        assert c.strides[1] == c.itemsize or c.size == 0

    _gpaw.rk(alpha, a, beta, c, trans)


def gpu_rk(alpha, a, beta, c, trans='c'):
    """Launch CPU or GPU version of rk()."""
    _gpaw.rk_gpu(alpha, a.data.ptr, a.shape,
                 beta, c.data.ptr, c.shape,
                 a.dtype)


def r2k(alpha, a, b, beta, c, trans='c'):
    """Rank-2k update of a matrix.

    Performs the operation::

                        dag        cc       dag
      c <- alpha * a . b    + alpha  * b . a    + beta * c

    or if trans='n'::
                    dag           cc   dag
      c <- alpha * a   . b + alpha  * b   . a + beta * c

    where ``a.b`` denotes the matrix multiplication defined by::

                 _
                \
      (a.b)   =  ) a         * b
           ij   /_  ipklm...     pjklm...
               pklm...

    ``cc`` denotes complex conjugation.

    ``dag`` denotes the hermitian conjugate (complex conjugation plus a
    swap of axis 0 and 1).

    Only the lower triangle of ``c`` will contain sensible numbers.
    """
    if debug:
        assert beta == 0.0 or is_finite(c, tril=True)
        assert (a.dtype == float and b.dtype == float and c.dtype == float or
                a.dtype == complex and b.dtype == complex and
                c.dtype == complex)
        assert a.flags.c_contiguous and b.flags.c_contiguous
        assert a.ndim > 1
        assert a.shape == b.shape
        if trans == 'c':
            assert c.shape == (a.shape[0], a.shape[0])
        else:
            assert c.shape == (a.shape[1], a.shape[1])
        assert c.strides[1] == c.itemsize or c.size == 0

    _gpaw.r2k(alpha, a, b, beta, c, trans)


def gpu_r2k(alpha, a, b, beta, c, trans='c'):
    """Launch CPU or GPU version of r2k()."""
    _gpaw.r2k_gpu(alpha, a.data.ptr, a.shape,
                  b.data.ptr, b.shape, beta,
                  c.data.ptr, c.shape,
                  a.dtype)


def gpu_dotc(a, b):
    r"""Dot product, conjugating the first vector with complex arguments.

    Returns the value of the operation::

        _
       \   cc
        ) a       * b
       /_  ijk...    ijk...
       ijk...

    ``cc`` denotes complex conjugation.
    """
    if debug:
        assert ((is_contiguous(a, float) and is_contiguous(b, float)) or
                (is_contiguous(a, complex) and is_contiguous(b, complex)))
        assert a.shape == b.shape

    return _gpaw.dotc_gpu(a.data.ptr, a.shape,
                          b.data.ptr, a.dtype)


def gpu_dotu(a, b):
    """Dot product, NOT conjugating the first vector with complex arguments.

    Returns the value of the operation::

        _
       \
        ) a       * b
       /_  ijk...    ijk...
       ijk...


    """
    if debug:
        assert ((is_contiguous(a, float) and is_contiguous(b, float)) or
                (is_contiguous(a, complex) and is_contiguous(b, complex)))
        assert a.shape == b.shape

    return _gpaw.dotu_gpu(a.data.ptr, a.shape,
                          b.data.ptr, a.dtype)


def _gemmdot(a, b, alpha=1.0, beta=1.0, out=None, trans='n'):
    """Matrix multiplication using gemm.

    return reference to out, where::

      out <- alpha * a . b + beta * out

    If out is None, a suitably sized zero array will be created.

    ``a.b`` denotes matrix multiplication, where the product-sum is
    over the last dimension of a, and either
    the first dimension of b (for trans='n'), or
    the last dimension of b (for trans='t' or 'c').

    If trans='c', the complex conjugate of b is used.
    """
    # Store original shapes
    ashape = a.shape
    bshape = b.shape

    # Vector-vector multiplication is handled by dotu
    if a.ndim == 1 and b.ndim == 1:
        assert out is None
        if trans == 'c':
            return alpha * np.vdot(b, a)  # dotc conjugates *first* argument
        else:
            return alpha * a.dot(b)

    # Map all arrays to 2D arrays
    a = a.reshape(-1, a.shape[-1])
    if trans == 'n':
        b = b.reshape(b.shape[0], -1)
        outshape = a.shape[0], b.shape[1]
    else:  # 't' or 'c'
        b = b.reshape(-1, b.shape[-1])

    # Apply BLAS gemm routine
    outshape = a.shape[0], b.shape[trans == 'n']
    if out is None:
        # (ATLAS can't handle uninitialized output array)
        out = np.zeros(outshape, a.dtype)
    else:
        out = out.reshape(outshape)
    mmmx(alpha, a, 'N', b, trans.upper(), beta, out)

    # Determine actual shape of result array
    if trans == 'n':
        outshape = ashape[:-1] + bshape[1:]
    else:  # 't' or 'c'
        outshape = ashape[:-1] + bshape[:-1]
    return out.reshape(outshape)


if not hasattr(_gpaw, 'mmm'):
    # These are the functions used with noblas=True
    # TODO: move these functions elsewhere so that
    # they can be used for unit tests

    def op(o, m):
        if o.upper() == 'N':
            return m
        if o.upper() == 'T':
            return m.T
        if o.upper() == 'C':
            return m.conj().T
        raise ValueError(f'unknown op: {o}')

    def rk(alpha, a, beta, c, trans='c'):  # noqa
        if c.size == 0:
            return
        if beta == 0:
            c[:] = 0.0
        else:
            c *= beta
        if trans == 'n':
            c += alpha * a.conj().T.dot(a)
        else:
            a = a.reshape((len(a), -1))
            c += alpha * a.dot(a.conj().T)

    def r2k(alpha, a, b, beta, c, trans='c'):  # noqa
        if c.size == 0:
            return
        if beta == 0.0:
            c[:] = 0.0
        else:
            c *= beta
        if trans == 'c':
            c += (alpha * a.reshape((len(a), -1))
                  .dot(b.reshape((len(b), -1)).conj().T) +
                  alpha * b.reshape((len(b), -1))
                  .dot(a.reshape((len(a), -1)).conj().T))
        else:
            c += alpha * (a.conj().T @ b + b.conj().T @ a)

    def mmm(alpha: T, a: np.ndarray, opa: str,  # noqa
            b: np.ndarray, opb: str,
            beta: T, c: np.ndarray) -> None:
        if beta == 0.0:
            c[:] = 0.0
        else:
            c *= beta
        c += alpha * op(opa, a).dot(op(opb, b))

    gemmdot = _gemmdot

elif not debug:
    mmm = _gpaw.mmm  # noqa
    rk = _gpaw.rk  # noqa
    r2k = _gpaw.r2k  # noqa
    gemmdot = _gemmdot

else:
    def gemmdot(a, b, alpha=1.0, beta=1.0, out=None, trans='n'):
        assert a.flags.c_contiguous
        assert b.flags.c_contiguous
        assert a.dtype == b.dtype
        if trans == 'n':
            assert a.shape[-1] == b.shape[0]
        else:
            assert a.shape[-1] == b.shape[-1]
        if out is not None:
            assert out.flags.c_contiguous
            assert a.dtype == out.dtype
            assert a.ndim > 1 or b.ndim > 1
            if trans == 'n':
                assert out.shape == a.shape[:-1] + b.shape[1:]
            else:
                assert out.shape == a.shape[:-1] + b.shape[:-1]
        return _gemmdot(a, b, alpha, beta, out, trans)



import sys
import functools
from time import time, sleep

from gpaw.utilities import devnull

from gpaw.utilities.memory import maxrss


class ProgressBar:
    def __init__(self, fd=sys.stdout, nobar=False):
        """Progress-bar.
        
        Usage::
            
            pb = ProgressBar()
            for i in range(10):
                pb.update(i / 10.0)
                do_stuff()
            pb.finish()
        """
        self.fd = fd
        
        try:
            self.tty = fd.isatty()
        except AttributeError:
            self.tty = False
        
        self.done = False
        self.n = None
        self.nobar = nobar
        self.t = time()
        self.symbols = ['-', '*']
        fd.flush()
        
    def update(self, x):
        """Update progress-bar (0 <= x <= 1)."""
        if x == 0 or self.done:
            return

        if self.tty:
            N = 35
        elif self.nobar:
            N = 10
        else:
            N = 40

        n = int(N * x)
        t = time() - self.t
        t_dt = self.format_time(t)
        est = self.format_time(t / x)
        p = functools.partial(print, file=self.fd)

        if self.tty:
            bar = '-' * (n - 1) + self.symbols[int(t % len(self.symbols))]
            p(('\r{0} / {1} ({2:.0f}%) |{3:' + str(N) + '}| ')
              .format(t_dt, est, x * 100, bar), end='')
            p(' {0:.0f} MB/core'.format(maxrss() / 1024**2), end='')
            if x == 1:
                p()
                self.done = True
            self.fd.flush()
        elif self.nobar:
            if self.n is None:
                p(('Started: {0:.0f} MB/core')
                  .format(maxrss() / 1024**2))
                self.n = 0
            if n > self.n:
                p(('{0} of {1} ({2:.0f}%) {3:.0f} MB/core')
                  .format(t_dt, est, x * 100, maxrss() / 1024**2))
                self.fd.flush()
                self.n = n
            if x == 1:
                p('Finished in {0}'.format(t_dt))
                self.fd.flush()
                self.done = True
        else:
            if self.n is None:
                p('{0}s |'.format(t / x), end='')
                self.n = 0
            if n > self.n:
                p('-' * (n - self.n), end='')
                self.fd.flush()
                self.n = n
            if x == 1:
                p('| Time: {0:.3f}s'.format(t))
                self.fd.flush()
                self.done = True
        
    def format_time(self, seconds):
        m, s = divmod(seconds, 60)
        h, m = divmod(m, 60)
        return '{0:.0f}h{1:.0f}m{2:.0f}s'.format(h, m, s)
                
    def finish(self):
        self.update(1)

    def enumerate(self, items):
        for i, item in enumerate(items):
            self.update(i / len(items))
            yield i, item
        self.finish()


def test():
    for fd in [sys.stdout, devnull, open('pb.txt', 'w')]:
        print(fd)
        pb = ProgressBar(fd)
        for i in range(20):
            pb.update(i / 20)
            sleep(0.03)
            pb.update((i + 1) / 20)
        pb.finish()


if __name__ == '__main__':
    test()



import numpy as np

from gpaw.grid_descriptor import GridDescriptor
from gpaw.spline import Spline
import gpaw.mpi as mpi
from gpaw.pw.descriptor import PWDescriptor
from gpaw.pw.lfc import PWLFC


def test_stress():
    x = 2.0
    rc = 3.5
    r = np.linspace(0, rc, 100)

    n = 40
    a = 8.0
    cell_cv = np.array([[a, 0.5, -1], [0, a, 2], [-1, 0, a + 1]])
    gd = GridDescriptor((n, n, n), cell_cv, comm=mpi.serial_comm)

    a_R = gd.empty()
    z = np.linspace(0, n, n, endpoint=False)
    a_R[:] = 2 + np.sin(2 * np.pi * z / n)

    spos_ac = np.array([(0.15, 0.45, 0.95)])

    pd = PWDescriptor(45, gd)
    a_G = pd.fft(a_R)

    s = Spline(0, rc, 2 * x**1.5 / np.pi * np.exp(-x * r**2))
    p = Spline(1, rc, 2 * x**1.5 / np.pi * np.exp(-x * r**2))
    d = Spline(2, rc, 2 * x**1.5 / np.pi * np.exp(-x * r**2))

    lfc = PWLFC([[s, p, d]], pd)
    lfc.set_positions(spos_ac)
    b_LG = pd.zeros(9)
    lfc.add(b_LG, {0: np.eye(9)})
    e1 = pd.integrate(a_G, b_LG)
    assert abs(lfc.integrate(a_G)[0] - e1).max() < 1e-11

    s1 = []
    for i in range(9):
        x = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        x[i] = 1
        s1.append(lfc.stress_tensor_contribution(a_G, {0: x}) -
                  np.eye(3) * e1[i])

    x = 1e-6
    for dist in [[[x, 0, 0], [0, 0, 0], [0, 0, 0]],
                 [[0, 0, 0], [0, x, 0], [0, 0, 0]],
                 [[0, 0, 0], [0, 0, 0], [0, 0, x]],
                 [[0, x, 0], [x, 0, 0], [0, 0, 0]],
                 [[0, 0, x], [0, 0, 0], [x, 0, 0]],
                 [[0, 0, 0], [0, 0, x], [0, x, 0]]]:
        e = dist + np.eye(3)
        c_cv = np.dot(cell_cv, e)
        gd = GridDescriptor((n, n, n), c_cv, comm=mpi.serial_comm)
        pd = PWDescriptor(45, gd)
        aa_G = a_G / np.linalg.det(e)
        lfc = PWLFC([[s, p, d]], pd)
        lfc.set_positions(spos_ac)
        b_LG = pd.zeros(9)
        lfc.add(b_LG, {0: np.eye(9)})
        e2 = pd.integrate(aa_G, b_LG)
        s2 = (e2 - e1) / x
        error = (np.array(s1) * dist).sum(1).sum(1) / x - s2
        print(s2, abs(error).max())
        assert abs(error).max() < 2e-6



import numpy as np
import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.fdtd.poisson_fdtd import FDTDPoissonSolver
from gpaw.fdtd.polarizable_material import (PermittivityPlus,
                                            PolarizableMaterial,
                                            PolarizableSphere)
from gpaw.mpi import world
from gpaw.tddft import TDDFT, DipoleMomentWriter
from gpaw.test import equal


@pytest.mark.later
def test_fdtd_ed(in_tmp_dir):
    # Accuracy
    energy_eps = 0.0005

    # Whole simulation cell (Angstroms)
    large_cell = [20, 20, 30]

    # Quantum subsystem
    atom_center = np.array([10.0, 10.0, 20.0])
    atoms = Atoms('Na2', [atom_center + [0.0, 0.0, -1.50],
                          atom_center + [0.0, 0.0, +1.50]])

    # Permittivity file
    if world.rank == 0:
        fo = open('ed.txt', 'w')
        fo.writelines(['1.20 0.20 25.0'])
        fo.close()
    world.barrier()

    # Classical subsystem
    classical_material = PolarizableMaterial()
    sphere_center = np.array([10.0, 10.0, 10.0])
    classical_material.add_component(
        PolarizableSphere(permittivity=PermittivityPlus('ed.txt'),
                          center=sphere_center,
                          radius=5.0))

    # Combined Poisson solver
    poissonsolver = FDTDPoissonSolver(classical_material=classical_material,
                                      qm_spacing=0.40,
                                      cl_spacing=0.40 * 4,
                                      cell=large_cell,
                                      remove_moments=(1, 4),
                                      communicator=world,
                                      potential_coupler='Refiner')
    poissonsolver.set_calculation_mode('iterate')

    # Combined system
    atoms.set_cell(large_cell)
    atoms, qm_spacing, gpts = poissonsolver.cut_cell(atoms,
                                                     vacuum=2.50)

    # Initialize GPAW
    gs_calc = GPAW(mode='fd',
                   gpts=gpts,
                   experimental={'niter_fixdensity': 2},
                   eigensolver='cg',
                   nbands=-1,
                   poissonsolver=poissonsolver,
                   symmetry={'point_group': False},
                   convergence={'energy': energy_eps})
    atoms.calc = gs_calc

    # Ground state
    energy = atoms.get_potential_energy()

    # Test ground state
    equal(energy, -0.631881, energy_eps * gs_calc.get_number_of_electrons())

    # Test floating point arithmetic errors
    equal(gs_calc.hamiltonian.poisson.shift_indices_1, [4, 4, 10], 0)
    equal(gs_calc.hamiltonian.poisson.shift_indices_2, [8, 8, 16], 0)

    # Save state
    gs_calc.write('gs.gpw', 'all')
    classical_material = None
    gs_calc = None

    # Initialize TDDFT and FDTD
    kick = [0.0, 0.0, 1.0e-3]
    time_step = 10.0
    max_time = 100  # 0.1 fs

    td_calc = TDDFT('gs.gpw')
    DipoleMomentWriter(td_calc, 'dm.dat')
    td_calc.absorption_kick(kick_strength=kick)

    # Propagate TDDFT and FDTD
    td_calc.propagate(time_step, max_time / time_step / 2)
    td_calc.write('td.gpw', 'all')

    td_calc2 = TDDFT('td.gpw')
    DipoleMomentWriter(td_calc2, 'dm.dat')
    td_calc2.propagate(time_step, max_time / time_step / 2)

    # Test
    ref_cl_dipole_moment = [5.25374117e-14, 5.75811267e-14, 3.08349334e-02]
    ref_qm_dipole_moment = [1.78620337e-11, -1.57782578e-11, 5.21368300e-01]

    tol = 1e-4
    equal(td_calc2.hamiltonian.poisson.get_classical_dipole_moment(),
          ref_cl_dipole_moment, tol)
    equal(td_calc2.hamiltonian.poisson.get_quantum_dipole_moment(),
          ref_qm_dipole_moment, tol)



import numpy as np
import pytest

from ase import Atoms

from gpaw import GPAW
from gpaw.mpi import world, serial_comm
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.lcaotddft.densitymatrix import DensityMatrix
from gpaw.lcaotddft.magneticmomentwriter import MagneticMomentWriter
from gpaw.lcaotddft.magneticmomentwriter import parse_header
from gpaw.tddft.spectrum import rotatory_strength_spectrum
from gpaw.tddft.units import as_to_au, eV_to_au, au_to_eV, rot_au_to_cgs

from gpaw.test import only_on_master
from . import parallel_options, check_txt_data, copy_and_cut_file

pytestmark = pytest.mark.usefixtures('module_tmp_path')

parallel_i = parallel_options()


@pytest.fixture(scope='module')
@only_on_master(world)
def initialize_system():
    comm = serial_comm

    atoms = Atoms('LiNaNaNa',
                  positions=[[0.0, 0.0, 0.0],
                             [2.0, 1.0, 0.0],
                             [4.0, 0.0, 1.0],
                             [6.0, -1.0, 0.0]])
    atoms.center(vacuum=4.0)

    calc = GPAW(nbands=2,
                h=0.4,
                setups={'Na': '1'},
                basis='sz(dzp)',
                mode='lcao',
                convergence={'density': 1e-12},
                communicator=comm,
                symmetry={'point_group': False},
                txt='gs.out')
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write('gs.gpw', mode='all')

    td_calc = LCAOTDDFT('gs.gpw',
                        communicator=comm,
                        txt='td.out')
    dmat = DensityMatrix(td_calc)
    MagneticMomentWriter(td_calc, 'mm.dat', dmat=dmat)
    MagneticMomentWriter(td_calc, 'mm_grid.dat', calculate_on_grid=True)
    MagneticMomentWriter(td_calc, 'mm_origin.dat',
                         origin='zero', origin_shift=[1.0, 2.0, 3.0])
    td_calc.absorption_kick([1e-5, 0., 0.])
    td_calc.propagate(100, 3)
    td_calc.write('td.gpw', mode='all')
    td_calc.propagate(100, 2)


@pytest.mark.rttddft
def test_magnetic_moment_values(initialize_system, module_tmp_path,
                                in_tmp_dir):
    with open('mm_ref.dat', 'w', encoding='utf-8') as f:
        f.write('''
# MagneticMomentWriter[version=4](origin='COM')
# origin_v = [7.634300, 5.000000, 4.302858] Å
#            time               mmx                    mmy                    mmz
          0.00000000     0.000000000000e+00     0.000000000000e+00     0.000000000000e+00
# Kick = [    1.000000000000e-05,     0.000000000000e+00,     0.000000000000e+00]; Time = 0.00000000
          0.00000000     8.192189793082e-06     1.038446327373e-05    -2.730498071751e-05
          4.13413733     7.838837723234e-06     1.000765310013e-05    -2.573300722038e-05
          8.26827467     6.809084660174e-06     8.879683492897e-06    -2.128890950807e-05
         12.40241200     5.175350632237e-06     7.009694921954e-06    -1.462938416394e-05
         16.53654934     3.058296873929e-06     4.443905967036e-06    -6.697375210691e-06
         20.67068667     6.247451722277e-07     1.298788405738e-06     1.460017881082e-06
'''.strip())  # noqa: E501

    check_txt_data(module_tmp_path / 'mm.dat', 'mm_ref.dat', atol=2e-14)


@pytest.mark.rttddft
def test_magnetic_moment_grid_evaluation(initialize_system, module_tmp_path):
    dpath = module_tmp_path
    check_txt_data(dpath / 'mm.dat', dpath / 'mm_grid.dat', atol=2e-8)


@pytest.mark.rttddft
@pytest.mark.parametrize('parallel', parallel_i)
def test_magnetic_moment_parallel(initialize_system, module_tmp_path, parallel,
                                  in_tmp_dir):
    td_calc = LCAOTDDFT(module_tmp_path / 'gs.gpw',
                        parallel=parallel,
                        txt='td.out')
    MagneticMomentWriter(td_calc, 'mm.dat')
    MagneticMomentWriter(td_calc, 'mm_grid.dat', calculate_on_grid=True)
    MagneticMomentWriter(td_calc, 'mm_origin.dat',
                         origin='zero', origin_shift=[1.0, 2.0, 3.0])
    td_calc.absorption_kick([1e-5, 0., 0.])
    td_calc.propagate(100, 5)

    for fname in ['mm.dat', 'mm_grid.dat', 'mm_origin.dat']:
        check_txt_data(module_tmp_path / fname, fname, atol=7e-14)


@pytest.mark.rttddft
@pytest.mark.parametrize('parallel', parallel_i)
def test_magnetic_moment_restart(initialize_system, module_tmp_path, parallel,
                                 in_tmp_dir):
    td_calc = LCAOTDDFT(module_tmp_path / 'td.gpw',
                        parallel=parallel,
                        txt='td.out')
    for fname in ['mm.dat', 'mm_grid.dat', 'mm_origin.dat']:
        if world.rank == 0:
            copy_and_cut_file(module_tmp_path / fname, fname, cut_lines=3)
        world.barrier()
        MagneticMomentWriter(td_calc, fname)
    td_calc.propagate(100, 2)

    for fname in ['mm.dat', 'mm_grid.dat', 'mm_origin.dat']:
        check_txt_data(module_tmp_path / fname, fname, atol=7e-14)


@only_on_master(world)
def test_spectrum(in_tmp_dir, rng):
    from gpaw.utilities.folder import Folder

    # Parameters for test data
    kick_strength = 1e-5
    frequency_v = np.array([1.0, 2.0, 3.0]) * eV_to_au
    strength_v = np.array([1.0, 2.0, 3.0])

    # Create dummy magnetic moment files
    for v, kick in enumerate('xyz'):
        kick_v = [0.0, 0.0, 0.0]
        kick_v[v] = kick_strength
        time_t = np.arange(0, 31e3, 10.0) * as_to_au
        data_tv = np.zeros((len(time_t), 4))
        data_tv[:, 0] = time_t
        # Fill unused columns with random values
        data_tv[:, 1:] = rng.random((len(time_t), 3))
        # Diagonal column has the data used for spectrum
        data_tv[:, v + 1] = (kick_strength * strength_v[v]
                             * np.cos(frequency_v[v] * time_t))
        with open(f'mm-{kick}.dat', 'w', encoding='utf-8') as f:
            f.write(f'''
# MagneticMomentWriter[version=4](origin='COM')
#            time               mmx                    mmy                    mmz
          0.00000000     0.000000000000e+00     0.000000000000e+00     0.000000000000e+00
# Kick = {kick_v}; Time = 0.00000000
''')  # noqa: E501
            np.savetxt(f, data_tv)

    # Calculate spectrum
    folding_kwargs = dict(folding='Gauss', width=0.2)
    rotatory_strength_spectrum(['mm-x.dat', 'mm-y.dat', 'mm-z.dat'],
                               'spec.dat',
                               **folding_kwargs,
                               e_min=0.0, e_max=5.0, delta_e=0.01)

    # Reference spectrum
    energy_e = np.arange(0, 5.0 + 1e-8, 0.01)
    f = Folder(**folding_kwargs).fold_values
    spec_e = (f(frequency_v * au_to_eV, strength_v, energy_e)[1]
              + f(-frequency_v * au_to_eV, strength_v, energy_e)[1])
    spec_e *= 0.5
    spec_e *= rot_au_to_cgs * 1e40

    # Compare spectrum
    data_ei = np.loadtxt('spec.dat')
    assert np.allclose(data_ei[:, 0], energy_e)
    assert np.allclose(data_ei[:, 1], spec_e)

    # Test failure
    with pytest.raises(RuntimeError):
        rotatory_strength_spectrum(['mm-x.dat', 'mm-x.dat', 'mm-z.dat'],
                                   'spec.dat')
    with pytest.raises(RuntimeError):
        rotatory_strength_spectrum(['mm-y.dat', 'mm-x.dat', 'mm-z.dat'],
                                   'spec.dat')
    with pytest.raises(RuntimeError):
        rotatory_strength_spectrum(['mm-x.dat', 'mm-y.dat', 'mm-x.dat'],
                                   'spec.dat')


def test_parse_header():
    line = 'SomeWriter[version=42](**{"str": "value", "vector": [1.2, 3.4], "boolean": true})'  # noqa: E501
    name, version, kwargs = parse_header(line)
    assert name == 'SomeWriter'
    assert version == 42
    assert kwargs['str'] == 'value'
    assert kwargs['vector'] == [1.2, 3.4]
    assert kwargs['boolean']

    # Test failure
    with pytest.raises(ValueError):
        parse_header('wrong line')
    with pytest.raises(ValueError):
        parse_header('A[version=1](**{wrong json})')
